/*
 * 4PSA VoipNow - Utility functions for constructing DB API queries
 *
 * Copyright (c) 2005-2009 Rack-Soft, LLC. All rights reserved.
 *
 * $Header: //itsd/devel/platform/voipnow/251/low_level/kamailio_extra_sources/1.5.4/db/db_utils_4psa.h#1 $
 *
 * Modified: $DateTime: 2010/09/28 18:16:02 $ by $Author: bogdanc $ with $Change: 71991 $
 */

/* This file provides a transparent way to work with Kamailio DB API for
 * performing different kind of queries. This is an attempt to move complicated
 * Kamailio implemented C pattern for performing DB Queries to a more simpler
 * and transparent approach.
 *
 * Sample Usage (SELECT):
 * - Declare the query based on an unique name
 * INIT_QUERY_WHERE(unique_query_name, max_where_column, max_results_column)
 *
 * - Add WHERE conditions to the query
 * ADD_QUERY_WHERE_COLUMN(unique_query_name, OP_EQ,  str_column_A,    STR, str_value);
 * ADD_QUERY_WHERE_COLUMN(unique_query_name, OP_LEQ, str_column_B,   INT, int_value);
 *  -> this will lead to a where condition of:
 *             WHERE (str_column_A == str_value) AND (str_column_B <= int_value)
 *
 * - Add the desired column to get in the select
 * ADD_QUERY_RESULT_COLUMN(unique_query_name, select_columnA)
 * ADD_QUERY_RESULT_COLUMN(unique_query_name, select_columnB)
 *  -> this will lead to a where condition of:
 *             SELECT select_columnA, select_columnB FROM table
 *
 * - Run the actual select using Kamailio DB API
 * 	 pa_dbf.query (pa_db, QUERY_COLS(unique_query_name),     QUERY_OPS(unique_query_name),
 *                        QUERY_VALS(unique_query_name),     QUERY_RESULTS(unique_query_name),
 *                        QUERY_COLS_CNT(unique_query_name), QUERY_RESULTS_CNT(unique_query_name),
 *                        0, &result)
 *
 * - When you will want to construct another query based on the same declared data
 * use RESET_QUERY(unique_query_name)
 *
 *
 * Files: modules/presence/presentity.c
 */


/* macros for setting a column */
#define ADD_QUERY_WHERE_COLUMN(query, op, col_name, type, value) \
	query##_ops[n_##query##_cols] = op; \
	ADD_QUERY_UPDATE_COLUMN(query, col_name, type, value);

#define ADD_QUERY_UPDATE_COLUMN(query, col_name, type, value) \
	query##_cols[n_##query##_cols] = col_name; \
	SET_##type##_VALUE(query##_vals[n_##query##_cols], value); \
	n_##query##_cols++;

#define ADD_QUERY_RESULT_COLUMN(query, col_name) \
	query##result_cols[n_##query##result_cols++] = col_name;

/* to start to redefine columns for a new query */
#define RESET_QUERY(query) \
	n_##query##result_cols = 0; \
	RESET_QUERY_UPDATE_VARS(query);

#define RESET_QUERY_UPDATE(query) \
	n_##query##_cols = 0;


/* macros for declaring a query type */
#define INIT_EMPTY_QUERY(query, no_results) \
	db_key_t query##result_cols[no_results]; \
	unsigned int n_##query##result_cols = 0;

#define INIT_QUERY_WHERE(query, no_cols, no_results) \
	db_op_t  query##_ops[no_cols]; \
	INIT_QUERY(query, no_cols, no_results);

#define INIT_QUERY(query, no_cols, no_results) \
	INIT_QUERY_UPDATE(query, no_cols); \
	INIT_EMPTY_QUERY(query, no_results);

#define INIT_QUERY_UPDATE_WHERE(query, no_cols) \
	db_op_t  query##_ops[no_cols]; \
	INIT_QUERY_UPDATE(query, no_cols);

#define INIT_QUERY_UPDATE(query, no_cols) \
	db_key_t query##_cols[no_cols]; \
	db_val_t query##_vals[no_cols]; \
	unsigned int n_##query##_cols = 0;


/* macros to get the needed columns in the DB API */
#define QUERY_COLS(query)        query##_cols
#define QUERY_VALS(query)        query##_vals
#define QUERY_OPS(query)         query##_ops
#define QUERY_RESULTS(query)     query##result_cols
#define QUERY_COLS_CNT(query)    n_##query##_cols
#define QUERY_RESULTS_CNT(query) n_##query##result_cols


/* macros for setting a column type value */
#define SET_STR_VALUE(_val, _str) \
	do{ \
		_val.type        = DB_STR; \
		_val.nul         = !(_str).s; \
		_val.val.str_val = _str; \
	} while(0);

#define SET_STRING_VALUE(_val, _string) \
	do{ \
		_val.type           = DB_STRING; \
		_val.nul            = !(_string); \
		_val.val.string_val = _string; \
	} while(0);

#define SET_RAW_VALUE(_val, _string) \
	do{ \
		_val.type           = DB_RAW; \
		_val.nul            = !(_string); \
		_val.val.string_val = _string; \
	} while(0);


#define SET_INT_VALUE(_val, _int) \
	do{ \
		_val.type        = DB_INT; \
		_val.nul         = 0; \
		_val.val.int_val = _int; \
	} while(0);

#define SET_BLOB_VALUE(_val, _blob) \
	do{ \
		_val.type        = DB_BLOB; \
		_val.nul         = !(_blob).s; \
		_val.val.blob_val = _blob; \
	} while(0);
