/*
 * Function for working with dynamic listen lists
 *
 * Copyright (c) 2005-2009 Rack-Soft, LLC. All rights reserved.
 *
 * $Header: //itsd/devel/platform/voipnow/251/low_level/kamailio_extra_sources/1.5.4/dynamic_sock_list_4psa.c#1 $
 *
 * Modified: $DateTime: 2010/09/28 18:16:02 $ by $Author: bogdanc $ with $Change: 71991 $
 */

#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/utsname.h>
#include <stdio.h>

#include <sys/ioctl.h>
#include <net/if.h>
#ifdef HAVE_SYS_SOCKIO_H
#include <sys/sockio.h>
#endif

#include "globals.h"
#include "socket_info.h"
#include "dprint.h"
#include "mem/mem.h"
#include "ut.h"
#include "resolve.h"
#include "name_alias.h"

extern int add_interfaces(char* if_name, int family, unsigned short port, unsigned short proto, struct socket_info** list);

static int *need_dynamic_socket_list = NULL;

static int is_addrany(struct socket_info *si)
{
	return !si->address.u.addr32[0];
}

static int check_need_dynamic_socket_list_local_4psa(int allow_compute)
{
	unsigned short proto = PROTO_UDP;
	struct socket_info **list;
	struct socket_info *si;
	int first_time = 0;
	
	if (need_dynamic_socket_list) {
		return *need_dynamic_socket_list;
	} else {
		if (!allow_compute) {
			/* If it is not the right time to check then say they are dynamic */
			return 1;
		}
		first_time = 1;
	}
	
	need_dynamic_socket_list = shm_malloc(sizeof(int));
	*need_dynamic_socket_list = 0;
		
	/* Check there is at least one 0.0.0.0 address present */
	do{
		list = get_sock_info_list(proto);
		
		for(si=list?*list:0; si; si=si->next){
			if (is_addrany(si)) {
				*need_dynamic_socket_list = 1;
				if (first_time) {
					LM_DBG("Dynamic socket lists enabled.");
				}
				return 1;
			}
		}
	
	}while((proto=next_proto(proto)));
	
	if (first_time) {
		LM_DBG("Dynamic socket lists not necessary.");
	}
	
	return 0;	
}

int have_dynamic_socket_list_4psa()
{
	return check_need_dynamic_socket_list_local_4psa(0);
}

int check_need_dynamic_socket_list_4psa()
{
	return check_need_dynamic_socket_list_local_4psa(1);
}

static char *loopback_str = "127.0.0.1";

static int is_public_address(struct socket_info *si)
{
	return !(si->flags & SI_IS_LO) && si->address.u.addr32[0] && memcmp(si->name.s, loopback_str, strlen(loopback_str));
}

/* Get a comma separated list of all socket names of the form
 * "proto1:ip1:port1, proto2:ip2:port2"
 * where ip is not loopback or 0.0.0.0
 */
char* get_socket_list_for_query_4psa(void)
{
	struct socket_info *si;
	struct socket_info** list;
	unsigned short proto;
	int string_size = 0;
	int current_size = 0;
	char *result = NULL;
	char *buf = NULL;
	int first = 1;
	
	LOCK_LISTEN_LIST
	
	proto = PROTO_UDP;
	/* Compute size of final string */
	do{
		list = get_sock_info_list(proto);
		for(si=list?*list:0; si; si=si->next){
			if (!is_public_address(si)) {
				continue;
			}
			string_size += si->sock_str.len + 4;
		}
	}while((proto=next_proto(proto)));

	if (string_size == 0) {
		LM_WARN("Nothing to add to socket list\n");
		goto _end;
	}
	
	if (!(result = pkg_malloc(string_size + 1))) {
		LM_ERR("Memory allocation failure\n");
		goto _end;
	}	
	
	/* Create socket list separated by commas */
	proto = PROTO_UDP;
	buf = result;
	first = 1;
	do{
		list = get_sock_info_list(proto);
		for(si=list?*list:0; si; si=si->next){
			if (!is_public_address(si)) {
				continue;
			}
			
			current_size = (buf - result) + si->sock_str.len;
			if (current_size > string_size) {
				LM_ERR("Socket list string size too small: %d < %d.\n", string_size, current_size);
				pkg_free(result);
				result = NULL;
				goto _end;
			}
			
			if (first) {
				first = 0;
				buf += sprintf(buf, "'%.*s'", si->sock_str.len, si->sock_str.s);
			} else {
				buf += sprintf(buf, ", '%.*s'", si->sock_str.len, si->sock_str.s);
			}
		}
	}while((proto=next_proto(proto)));
	/* Add terminating NULL character */
	buf[0] = 0;
	
_end:
	UNLOCK_LISTEN_LIST
	return result;	
}


static int socket_info_matches(struct socket_info *left, struct socket_info *right)
{
	return (left->port_no == right->port_no) && (left->address.af == right->address.af) && !memcmp(left->address.u.addr, right->address.u.addr, left->address.len);
}

static struct socket_info *find_matching_si(struct socket_info *si_needed, struct socket_info **list)
{
	struct socket_info *si;
	
	for(si=list?*list:0; si; si=si->next){
		if (socket_info_matches(si,si_needed)) {
			return si;
		}
	}
	
	return NULL;
}

static void print_socket_list_4psa(const char *name, struct socket_info** list)
{
	struct socket_info *si;

	LM_DBG("<======== Socket list[%s]:\n", name);	
	for(si=list?*list:0; si; si=si->next){
		LM_DBG("\t\t%s: %s [%s]:%s%s\n", get_proto_name(si->proto),
				si->name.s, si->address_str.s, si->port_no_str.s, 
				si->flags & SI_IS_MCAST ? " mcast" : "");
	}
	LM_DBG("========>\n");	
}

int fix_socket_info_4psa(struct socket_info * si) {
	char* tmp;
	int len;
	struct hostent* he;
	char** h;

	/* fix port number, port_no should be !=0 here */
	if (si->port_no==0){
#ifdef USE_TLS
		si->port_no= (si->proto==PROTO_TLS)?tls_port_no:port_no;
#else
		si->port_no= port_no;
#endif
	}
	tmp=int2str(si->port_no, &len);
	if (len>=MAX_PORT_LEN){
		LM_ERR("bad port number: %d\n", si->port_no);
		goto error;
	}
	si->port_no_str.s=(char*)shm_malloc(len+1);
	if (si->port_no_str.s==0){
		LM_ERR("out of pkg memory.\n");
		goto error;
	}
	strncpy(si->port_no_str.s, tmp, len+1);
	si->port_no_str.len=len;
	/* get "official hostnames", all the aliases etc. */
	he=resolvehost(si->name.s,0);
	if (he==0){
		LM_ERR("could not resolve %s\n", si->name.s);
		goto error;
	}
	/* check if we got the official name */
	if (strcasecmp(he->h_name, si->name.s)!=0){
		if (auto_aliases && add_alias(si->name.s, si->name.len,
					si->port_no, si->proto)<0){
			LM_ERR("add_alias failed\n");
		}
		/* change the official name */
		shm_free(si->name.s);
		si->name.s=(char*)shm_malloc(strlen(he->h_name)+1);
		if (si->name.s==0){
			LM_ERR("out of pkg memory.\n");
			goto error;
		}
		si->name.len=strlen(he->h_name);
		strncpy(si->name.s, he->h_name, si->name.len+1);
	}
	/* add the aliases*/
	if (auto_aliases) {
		for(h=he->h_aliases; h && *h; h++)
			if (add_alias(*h, strlen(*h), si->port_no, si->proto)<0){
				LM_ERR("add_alias failed\n");
			}
	}
	hostent2ip_addr(&si->address, he, 0); /*convert to ip_addr
											format*/
	if ((tmp=ip_addr2a(&si->address))==0) goto error;
	if (si->address.af == AF_INET6) {
		si->address_str.s=(char*)shm_malloc(strlen(tmp)+1+2);
		if (si->address_str.s==0){
			LM_ERR("out of pkg memory.\n");
			goto error;
		}
		si->address_str.s[0] = '[';
		strncpy( si->address_str.s+1 , tmp, strlen(tmp));
		si->address_str.s[1+strlen(tmp)] = ']';
		si->address_str.s[2+strlen(tmp)] = '\0';
		si->address_str.len=strlen(tmp) + 2;
	} else {
		si->address_str.s=(char*)shm_malloc(strlen(tmp)+1);
		if (si->address_str.s==0){
			LM_ERR("out of pkg memory.\n");
			goto error;
		}
		strncpy(si->address_str.s, tmp, strlen(tmp)+1);
		si->address_str.len=strlen(tmp);
	}
	/* set is_ip (1 if name is an ip address, 0 otherwise) */
	if ( auto_aliases && (si->address_str.len==si->name.len) &&
			(strncasecmp(si->address_str.s, si->name.s,
						 si->address_str.len)==0)
	   ){
		si->flags|=SI_IS_IP;
		/* do rev. DNS on it (for aliases)*/
		he=rev_resolvehost(&si->address);
		if (he==0){
			LM_WARN("could not rev. resolve %s\n", si->name.s);
		}else{
			/* add the aliases*/
			if (add_alias(he->h_name, strlen(he->h_name),
						si->port_no, si->proto)<0){
				LM_ERR("add_alias failed\n");
			}
			for(h=he->h_aliases; h && *h; h++)
				if (add_alias(*h,strlen(*h),si->port_no,si->proto)<0){
					LM_ERR(" add_alias failed\n");
				}
		}
	}

	/* build and set string encoding */
	tmp = socket2str( si, 0, &si->sock_str.len);
	if (tmp==0) {
		LM_ERR("failed to convert socket to string");
		goto error;
	}
	si->sock_str.s=(char*)shm_malloc(si->sock_str.len + 1);
	if (si->sock_str.s==0) {
		LM_ERR("out of pkg memory.\n");
		goto error;
	}
	memcpy(si->sock_str.s, tmp, si->sock_str.len);

#ifdef USE_MCAST
	/* Check if it is an multicast address and
	 *          * set the flag if so
	 *                   */
	if (is_mcast(&si->address)) {
		si->flags |= SI_IS_MCAST;
	}
#endif /* USE_MCAST */

	return 0;
error:
	return -1;
}

static int fix_addrany_siblings_4psa(struct socket_info *any, struct socket_info **list, struct socket_info **cache_list)
{
	struct socket_info *tmp_list = NULL;
	struct socket_info *si;
	struct socket_info *new_si;
	struct socket_info *cached_si;
	int newly_added = 0;
	struct socket_info *next;
		
	LM_DBG("Verifying interface changes for %.*s.\n", any->sock_str.len, any->sock_str.s);
	
	if (add_interfaces(0, AF_INET, any->port_no, any->proto, &tmp_list) < 0) {
		LM_ERR("Failed to get socket list.\n");
		goto _end;
	}
	
	for(new_si=tmp_list; new_si; new_si = new_si->next){
		fix_socket_info_4psa(new_si);
	}
	
	print_socket_list_4psa("Current interface list", list);
	print_socket_list_4psa("New interface list", &tmp_list);
	print_socket_list_4psa("Cached interface list", cache_list);
	
	/* Remove interfaces that are no longer present and move them to cache 
	 * They are not freed as thet might be in use.
	 */
	for(si=list?*list:0; si;){
		next = si->next;
		if (si->parent == any) {
			if (is_public_address(si) && !find_matching_si(si, &tmp_list)) {
				LM_DBG("Interface %.*s removed.\n", si->sock_str.len, si->sock_str.s);
			
				si->flags |= SI_IS_DELETED;
				
				/* Move it to cache */
				sock_listrm(list, si);
				si->prev = NULL;
				si->next = NULL;
				sock_listadd(cache_list, si);	
			}
		}
		si = next;	
	}
	
	/* Add interfaces that are in tmp_list but not in the listen list. 
	 * If they are in cache use those.
	 */
	for(new_si=tmp_list; new_si;){
		next = new_si->next;
		if (is_public_address(new_si) && !find_matching_si(new_si, list)) {
			/* Search in cache */
			if ((cached_si = find_matching_si(new_si, cache_list))) {
				LM_DBG("Interface %.*s added from cache.\n", new_si->sock_str.len, new_si->sock_str.s);
				
				/* Move from cache list to listen list */
				sock_listrm(cache_list, cached_si);
				cached_si->prev = NULL;
				cached_si->next = NULL;
				sock_listadd(list, cached_si);		
				
				cached_si->flags &= ~SI_IS_DELETED;
				new_si = cached_si;
			} else {
				LM_DBG("Interface %.*s newly added.\n", new_si->sock_str.len, new_si->sock_str.s);
				
				/* Move from new list to listen list */
				sock_listrm(&tmp_list, new_si);
				new_si->prev = NULL;
				new_si->next = NULL;
				sock_listadd(list, new_si);	
			}
			
			newly_added++;
			
			/* Clone important information */
			new_si->port_no = any->port_no; /* fix port number if changed */
			new_si->socket = any->socket;
			new_si->su = any->su;
			new_si->su.sin.sin_addr.s_addr = new_si->address.u.addr[0]; /* Only IPv4 for now */
			new_si->parent = any;
		}
		new_si = next;	
	}
	
	print_socket_list_4psa("Current interface list(DONE)", list);
	print_socket_list_4psa("New interface list(DONE)", &tmp_list);
	print_socket_list_4psa("Cached interface list(DONE)", cache_list);
			
	/* Cleanup unused tmp_list entries */
	for(new_si=tmp_list; new_si;){
		next = new_si->next;
		free_sock_info(new_si);
		new_si = next;
	}

_end:
	return newly_added;	
}

struct socket_info *find_address_4psa( struct ip_addr *addr, struct socket_info *any)
{
	struct socket_info *result = NULL;
	struct socket_info *si = NULL;
	struct socket_info **list;
	struct socket_info **cache_list;
	int found = 0;
	char *addr_str;
	
	LOCK_LISTEN_LIST
		
	list = get_sock_info_list(any->proto);
	
	if (!list) {
		goto _end;
	}
	
	for(si=list?*list:0; si; si = si->next){
		if ((si->parent == any) && !memcmp(addr->u.addr, si->address.u.addr, addr->len)) {
			result = si;
			found = 1;
			goto _end;
		}
	}
	
	cache_list = get_sock_info_cache_list(any->proto);
	
	/* None found so force a recheck of public interfaces */
	fix_addrany_siblings_4psa(any, list, cache_list);
	
	/* Check if one matching was added. If none is found take last public IP as address */
	for(si=list?*list:0; si; si = si->next)
	{
		if ((si->parent == any) && !memcmp(addr->u.addr, si->address.u.addr, addr->len)) {
				result = si;
				found = 1;
				goto _end;
		}
	}
	
_end:	
	UNLOCK_LISTEN_LIST
		
	if (found) {
		LM_DBG("Found match in socket list for address %.*s.", si->sock_str.len, si->sock_str.s);
	} else {
		addr_str = ip_addr2a(addr);
		if (!addr_str) {
			addr_str = "<unknown>";
		}
		result = any;
		LM_WARN("Picked %.*s because we could not find interface %s in list.", result->sock_str.len, result->sock_str.s, addr_str);
	}
	return result;
}

/* choose the first non 0.0.0.0, public interface address from the listen addresses */
struct socket_info* choose_public_address_4psa(struct socket_info* list, struct socket_info* bind_address)
{
	struct socket_info *si;
	struct socket_info *result = NULL;
	struct socket_info *any = NULL;
		
	if (bind_address && !bind_address->parent && is_addrany(bind_address)) {
		any = bind_address;
	}
	
	LOCK_LISTEN_LIST
	
	for(si=list; si; si=si->next) {
		if (bind_address && bind_address->port_no != si->port_no) {
			continue;
		}
		
		if (any && (si->parent != any)) {
			continue;
		}
		
		if (is_public_address(si)) {
			result = si;
			break;
		}
	}
	
	UNLOCK_LISTEN_LIST	

	if (result) {
		LM_DBG("Picked public address %.*s.", result->sock_str.len, result->sock_str.s);
	} else {
		if (bind_address) {
			LM_DBG("Picked binding address %.*s.", bind_address->sock_str.len, bind_address->sock_str.s);
		} else {
			LM_WARN("Could not pick any address.");
		}
	}
	
	return result;
}

/* For all addresses of the form 0.0.0.0:port 
 * add all new public addresses or remove those 
 * that were removed from the system
 */
void update_addresses_4psa(void)
{
	unsigned short proto = PROTO_UDP;
	struct socket_info **list;
	struct socket_info **cache_list;
	struct socket_info *si;

	if (!check_need_dynamic_socket_list_4psa())	
		return;		
	
	LOCK_LISTEN_LIST
	
	do{
		list = get_sock_info_list(proto);
		cache_list = get_sock_info_cache_list(proto);
		
		if (list) {
			
			/* For all addresses of form 0.0.0.0 fix their siblings */
			for(si=*list; si; si=si->next){
				if (is_addrany(si)) {
					fix_addrany_siblings_4psa(si, list, cache_list);
				}
			}
		}
	
	}while((proto=next_proto(proto)));
	
	UNLOCK_LISTEN_LIST	
}
 
