/*
 * 4PSA VoipNow - GLIBC weak_alias malloc OVERWRITES.
 *                We are defining our own implementation of the default system
 *                allocator(malloc) that is used in both Glibc functions (such
 *                as strdup) or in other Kamailio code (such as byson files) or
 *                in 3rd party library code (such as libmemcached).
 *                This is possible since Linux GLIBC uses weak_alias linker flag
 *                for defining the malloc API which enables the loader to overwrite
 *                the functions with the first one defined, without generating an error.
 *                For this purpose we add the malloc API definitions into the elf itself
 *
 * Copyright (c) 2005-2009 Rack-Soft, LLC. All rights reserved.
 *
 * $Header: //itsd/devel/platform/voipnow/251/low_level/kamailio_extra_sources/1.5.4/mem/malloc_wrapper.c#1 $
 *
 * Modified: $DateTime: 2010/09/28 18:16:02 $ by $Author: bogdanc $ with $Change: 71991 $
 */

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mem.h"
#include "../dprint.h"


/* ================== GLIBC weak_alias malloc OVERWRITES =================== */
#ifdef SLAB_MALLOC
/* Note*: Only works with private memory slab allocator
 * implementation. The kamailio default allocators must
 * be modified in order to work with libc.
 * make sure you do not syslog here or you'll get
 * an infinite recursion as syslog API uses malloc
 */
void* malloc(size_t size)
{
	return pkg_malloc(size);
}

void* calloc(size_t nmemb, size_t size)
{
	void *ptr = NULL;

	size_t total_size = nmemb * size;
	ptr = malloc(total_size);
	memset(ptr, 0, total_size);
	return ptr;
}

void free(void *ptr)
{
	return pkg_free(ptr);
}

void* realloc(void *ptr, size_t size)
{
	return pkg_realloc(ptr, size);
}
/* ========================================================================= */



/* ================= GLIBC weak_alias obsolete OVERWRITES ================== */
#define MAX_TRY_ALIGNEMENT 254 /* the maximum retries for alignment */
static void* try_align[MAX_TRY_ALIGNEMENT + 2]; /* not thread safe  */
static long kPageSize = 0;     /* the current page size (4k)        */

/* The function posix_memalign() allocates size bytes and places the address
 * of the allocated memory in *memptr. The address of the allocated memory
 * will be a multiple of alignment, which must be a power of two and
 * a multiple of sizeof(void *)
 */
int posix_memalign(void **memptr, size_t alignment, size_t size)
{
	int i    = 0;
	size_t p = 0;

	/* try to allocate until an alignment pointer is received */
	do {
		if (!(try_align[i] = pkg_malloc(size))) {
			break;
		}
		p = (size_t)try_align[i];
		try_align[++i] = (void*)0;
	} while (((p % alignment) != 0) && i < MAX_TRY_ALIGNEMENT);

	if (i == MAX_TRY_ALIGNEMENT)
		errno = ENOMEM;

	i = 0;
	/* free all previously allocated pointers */
	while (try_align[i] && try_align[i + 1]) {
		pkg_free(try_align[i++]);
	}

	if (errno == ENOMEM) {
		if (try_align[i]) pkg_free(try_align[i]);
		*memptr = (void*)0;
		return ENOMEM;
	}

	*memptr = try_align[i];
	return 0;
}

/* The obsolete function memalign() allocates size bytes and returns
 * a pointer to the allocated memory. The memory address will be a
 * multiple of boundary, which must be a power of two
 */
void* memalign(size_t boundary, size_t size)
{
	void *p = (void*)0;
	posix_memalign(&p, boundary, size);
	return p;
}

/* The  obsolete  function  valloc() allocates size bytes and returns
 * a pointer to the allocated memory.  The memory address will be a
 * multiple of the page size.  It is equivalent to
 * memalign(sysconf(_SC_PAGESIZE),size)
 */
void* valloc(size_t size)
{
	if (!kPageSize) {
		 kPageSize = sysconf(_SC_PAGESIZE);
	}
	return memalign(kPageSize, size);
}
#endif
/* ========================================================================= */
