/*
 * 4PSA VoipNow - Implements a custom memory allocator for shared memory
 *                allocation based on the private memory SLAB allocator.
 *                Pratically the allocation logic is unchanged, the only
 *                difference is in the allocation of shared memory segments.
 *                We allocate a SLAB to a SLAB class from a previously mapped
 *                shared memory. The allocation of the shared memory is lazy
 *                and is done on a demand basis (when shm_slab_malloc needs
 *                additional shared memory) or on a SIGFAULT signal.
 *                Valgrind debugging is also supported.
 *
 * Copyright (c) 2009 Rack-Soft, LLC. All rights reserved..
 *
 * $Header: //itsd/devel/platform/voipnow/251/low_level/kamailio_extra_sources/1.5.4/mem/shm_slab_malloc.c#1 $
 *
 * Modified: $DateTime: 2010/09/28 18:16:02 $ by $Author: bogdanc $ with $Change: 71991 $
 */

#ifdef SHM_SLAB_MALLOC

#include <signal.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include "shm_slab_malloc.h"
#include "../globals.h"
#include "../statistics.h"
#include "../ut.h"


/* ================================= Forward Declarations ================================ */
static inline int    shm_update_mapped    (void);
static        void   shm_sigsegv_signal   (int sigNo);
static        void   shm_sigsegv_sigaction(int sigNo, siginfo_t* sig_info, void* context);
/* ======================================================================================= */


/* =============================== Internal Types and Data =============================== */
struct _slabclass;
/* internal SLAB allocation data, stored in shared memory for a consistent view accross all processes */
typedef struct _shm_slab {
	int power_largest;
	struct _slabclass* slabclass;

	/* internal variables for mapping private memory */
	void* shm_new_slab; /* where mapped area starts from which to allocate a new SLAB  */
	void* shm_end_slab; /* where mapped area ends and we need to map additional memory */
} shm_slab_t;

const  char*            shm_mem_obj_dev = "kamailio_shm_slab"; /* the shared memory object name */
static int              shm_mem_obj;  /* the shared object for mapping the shared regions       */
static void*            prv_end_slab; /* where shared mapped area ends in the current process   */
static shm_slab_t*      shm;          /* the SLAB allocator's data stored in shared memory      */
static struct sigaction libcSIGSEGV;  /* old libc's signal handler for SIGSEGV signal           */
/* ======================================================================================= */


/* ======================== Shared memory SLAB specifics definitions ===================== */
#define SLAB_MAP_START SLAB_SHARED_MAP_START

#define GET_SLAB_CLASS(class_id) (&shm->slabclass[class_id])

#define POWER_LAGEST shm->power_largest

#define SLAB_MALLOC_F  shm_slabs_malloc
#define SLAB_FREE_F    shm_slabs_free
#define SLAB_REALLOC_F shm_slabs_realloc
#define SLAB_INFO_F    shm_slabs_info
#define SLAB_STATUS_F  shm_slabs_status

/* change the default SLAB size for larger shm allocs */
#if   defined(__CPU_i386)
	#define POWER_BLOCK 2097152 /* 2MB */
#elif defined(__CPU_x86_64)
	#define POWER_BLOCK 3145728 /* 3MB */
#else
	#define POWER_BLOCK 1048576 /* 1MB */
#endif

static size_t SLAB_MAGIC = 0xDDDD; /* D stands for shareD */

#define MIN_FTRUNCATE 1048576 /* minimum chunk size ftruncate will try - 1MB */

/* we also need to define a shared valgrind if valgrind will be used */
#define __VALGRIND_SHARED__ 1
/* now we can safetly include the SLAB ALLOCATOR API */
#include "slab_allocator.h"
/* ======================================================================================= */


/* ================= Internal Macros for Private memory SLAB allocation ================== */
static inline int ftruncate_4psa(off_t file_size, int to_add)
{
	int remaining = to_add;
	int no_tries  = 0;

	do {
		if ( ftruncate(shm_mem_obj, file_size + (off_t)to_add) == -1 ) {
			if (!no_tries) {
				no_tries = 1;
				/* try to reopen the shared memory object */
				if ((shm_mem_obj = shm_open(shm_mem_obj_dev, O_RDWR,
							S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP)) >= 0 ) {
					continue;
				}
			} else {
				/* try to fix an error where on some machines a tmpfs filesystem can be resized only in small chunks */
				to_add >>= 1;
				if (to_add > MIN_FTRUNCATE) {
					LM_WARN("Trying ftruncate with %d MB\n", to_add >> 20);
					continue;
				}
			}

			LM_CRIT("Could not add additional shared memory: ftruncate failed: offset=%lld , error=%s\n", (long long)file_size, strerror(errno));
			return -1;
		}
		file_size += to_add;
		remaining -= to_add;
	} while (remaining > 0);

	return 0;
}


static inline void* internal_malloc(const size_t size)
{
	void *p = (void*)0;

	if (prv_end_slab < shm->shm_end_slab) {
		shm_update_mapped(); /* map all previously mapped shared memory */
	}

	if ((size_t)(shm->shm_end_slab - shm->shm_new_slab) < size) {
		void *new_chunk = (void*)0;
		off_t offset    = 0;

		offset = (off_t)prv_end_slab - SLAB_SHARED_MAP_START;
		/* modify the shared memory object size to support the new mapped area */
		if ( ftruncate_4psa(offset, MAP_PREALLOC) == -1 ) {
			return p;
		}

		/* we need to map additional memory. We do it at a succesize address */
		if ( ((new_chunk = mmap(prv_end_slab, MAP_PREALLOC, PROT_READ | PROT_WRITE,
							MAP_SHARED | MAP_FIXED, shm_mem_obj, offset)) == (void*)-1)
				|| (new_chunk != prv_end_slab) ) {
			LM_CRIT("Could not add additional shared memory: mmap failed: %s\n", strerror(errno));
			if ( (new_chunk != (void*)-1) && (munmap(new_chunk, MAP_PREALLOC) == -1) ) {
				LM_ERR("Failed to unmap shared memory at the wrong address: %s\n", strerror(errno));
			}
		} else {
			p                 = shm->shm_new_slab;
			shm->shm_new_slab = (char*)shm->shm_new_slab + size;
			prv_end_slab      = (char*)prv_end_slab + MAP_PREALLOC;
			shm->shm_end_slab = prv_end_slab;
			LM_WARN("Increased shared memory with %lu MB from %p to %p",
					(unsigned long int)MAP_PREALLOC >> 20,
					(void*)(size_t)SLAB_SHARED_MAP_START,
					(void*)shm->shm_end_slab);
		}
	} else {
		p                 = shm->shm_new_slab;
		shm->shm_new_slab = (char*)shm->shm_new_slab + size;
	}

	return p;
}


static inline void  internal_free(void *p, const size_t size)
{
	/* We currently do not support unmaping SLABs from memory */
}

static inline void* internal_realloc(void *old_p, const size_t old_size, const size_t new_size)
{
	void *new_p = internal_malloc(new_size);
	if (new_p && old_p) {
		memcpy(new_p, old_p, old_size);
		internal_free(old_p, old_size);
	}
	return new_p;
}
/* ======================================================================================= */


/* ============================== PUBLIC API implementation ============================== */
/**
 * \brief Shared memory SLAB allocator system initialization
 *
 * Prior to calling the default initialization for the SLAB classes of a
 * SLAB allocator we initialize the variables for current range of the
 * available shared memory mapped, in the private member of each process.
 * The range is currently 0 and starts at SLAB_SHARED_MAP_START. We also
 * initialize a shared memory object used to map additional shared memory
 * in each process, and allocate a first chunk in order to place the shm
 * shared SLAB structures in it, to have a consistent view of the allocation/
 * deallocation across all processes. Also we add a SIGSEGV signal handler
 * to catch unmapped shared memory in each process and map it in that process
 * if needed.
 *
 * \param factor the factor used to grow the SLOT size in each new SLAB class
 * \return 0/-1 on SUCCESS/FAIL
 */
int shm_slabs_init(const double factor)
{
	struct sigaction mySIGSEGV;
	int tries = 0;
	int uid = -1; /* -1 means will be ignored by fchown */
	int gid = -1;

	shm = (void*)0;
	
	/* user and group defined in globals.h represent the runtime user of Kamailio */
	if (user && (user2uid(&uid, &gid, user) < 0)) {
		LM_ERR("bad user name/uid number: -u %s\n", user);
		uid = -1;
	}
	if (group && (group2gid(&gid, group) < 0)) {
		LM_ERR("bad group name/gid number: -u %s\n", group);
		gid = -1;
	}

try_again:
	if ( (shm_mem_obj = shm_open(shm_mem_obj_dev, O_RDWR | O_CREAT | O_EXCL,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP)) == -1 ) {
		if (!tries++ && !shm_unlink(shm_mem_obj_dev))
			goto try_again;
		LM_CRIT("Could not open a shared memory object. No shared memory will be available."
				" shm_open: %s\n", strerror(errno));
		goto error;
	}
	
	/* try to change current ownership of the file descriptor */
	if (fchown(shm_mem_obj, uid, gid) < 0) {
		LM_CRIT("Could not change the permissions of the shared memory file. Continuing anyway"
				" fchown: %s\n", strerror(errno));
	}

	if ( ftruncate_4psa((off_t)0, MAP_PREALLOC) == -1 ) {
		LM_CRIT("Could not resize the shared memory object. No shared memory will be available\n");
		goto error;
	}

	/* We are setting the shared memory start address */
	prv_end_slab = (void*)(size_t)SLAB_SHARED_MAP_START;

	/* alloc a initial zone in shared memory to set the SLAB allocator shared data structure */
	if ( (shm = (shm_slab_t*)mmap(prv_end_slab, MAP_PREALLOC, PROT_READ | PROT_WRITE, \
					MAP_SHARED | MAP_FIXED, shm_mem_obj, 0)) == (shm_slab_t*)-1 ) {
		LM_CRIT("Could not add additional shared memory: mmap failed: %s\n", strerror(errno));
		goto error;
	}
	memset(shm, 0, sizeof(shm_slab_t) + sizeof(slabclass_t) * MAX_NUMBER_OF_SLAB_CLASSES);
	shm->slabclass = (slabclass_t*)((char*)shm + sizeof(shm_slab_t));

	/* we set the new address from where we can start adding SLABs */
	shm->shm_new_slab = (void*)ROUNDTO_PAGE_SIZE((char*)shm + sizeof(shm_slab_t) + \
			sizeof(slabclass_t) * MAX_NUMBER_OF_SLAB_CLASSES);
#ifdef __VALGRIND__
	/* now we should init the valgrinds operation */
	shm->shm_new_slab = shared_valgrind_init(shm->shm_new_slab);
#endif
	shm->shm_end_slab = prv_end_slab = (char*)shm + MAP_PREALLOC;

	/* we call the private slab allocation api to initialize the slabs */
	init_slabs(factor);

	/* install SHM_MALLOC SIGSEGV signal handler for sync shared mappings
	 * across processes on a lazy per demand basis
	 */
	memset(&mySIGSEGV, 0, sizeof(struct sigaction));
	sigemptyset(&mySIGSEGV.sa_mask);     /* all signals unblocked     */
	mySIGSEGV.sa_handler   = shm_sigsegv_signal;
	mySIGSEGV.sa_sigaction = shm_sigsegv_sigaction;
	/* force to use sa_sigaction and not interruptable functions to restart*/
	mySIGSEGV.sa_flags     = SA_SIGINFO | SA_RESTART;

	 /* set up SIGSEGV signal handler */
	if (sigaction(SIGSEGV, &mySIGSEGV, &libcSIGSEGV) == -1) {
		LM_ERR("SHM SLAB MALLOC: no SIGSEGV signal handler can be installed."
				" No shared memory will be available \n");
		goto error;
	}

	return 0;
error:
	return -1;
}


#ifdef STATISTICS
/**
 * \brief Shared memory SLAB allocator statistics
 *
 * Used by Kamailio statistic API to get different type
 * of information for the used space.
 *
 * \param type the type of the statistic to be returned
 * \return the available statistic information
 */
unsigned long shm_slab_get(shm_slab_statistics_t type)
{
	struct mem_info info;
	shm_slabs_info(&info);
	switch (type) {
		case SIZE:  return info.total_size;
		case MUSED:
		case USED:  return info.max_used;
		case RUSED: return info.real_used;
		case FREE:  return info.free;
		case FRAGS: return info.total_frags;
	}
	return info.total_size;
}
#endif
/* ======================================================================================= */


/* ============================= Internal API implementation ============================= */
/**
 * \brief Kamailio implemented functionality for synchronising the allocated
 * shared memory in this process.
 *
 * Called when a SIGFAULT signal is raised tries to allocate all shared memory
 * regions that have been previously mapped in other processes
 *
 * \return 0/1 if no map or new map was added
 */
static inline int shm_update_mapped(void)
{
	off_t offset = 0;
	size_t size  = 0;

	if (shm->shm_end_slab <= prv_end_slab)
		return 0;

	/* compute the shared memory size and memory object offset to map */
	size   = (size_t)shm->shm_end_slab - (size_t)prv_end_slab;
	offset = (off_t)prv_end_slab - SLAB_SHARED_MAP_START;

	if ( ftruncate_4psa(offset, (int)size) == -1 ) {
		LM_CRIT("Could not resize the shared memory object. No shared memory will be available\n");
		return 0;
	}

	/* synchronize the allocated shared memory in this process */
	if ( (prv_end_slab = mmap(prv_end_slab, size, PROT_READ | PROT_WRITE, \
					MAP_SHARED | MAP_FIXED, shm_mem_obj, offset)) == (void*)-1 ) {
		LM_CRIT("Could not synchronize shared memory: mmap failed: %s\n", strerror(errno));
		prv_end_slab = (void*)0;
		return 0;
	}
	prv_end_slab = shm->shm_end_slab;
	/* mapped a new area */
	LM_WARN("Updated shared memory with %lu MB to %p",
				(unsigned long int)size >> 20,
				(void*)prv_end_slab);
	return 1;
}


/**
 * \brief SIGSEGV signal handler for shared memory misses
 * shared memory in this process.
 *
 * Overwritten SIGFAULT libc's handler. Tries to map the unmapped
 * shared memory, or will fail and call default libc's handler if
 * no new shared memory is mapped.
 * This is used only if the current system signal implementation
 * does not support the Realtime Signals Exception option or the
 * XSI Extension option.
 *
 * \param sigNo the signal id
 * \return nothing
 */
static void shm_sigsegv_signal(int sigNo)
{
	shm_sigsegv_sigaction(sigNo, 0, 0);
}


/**
 * \brief SIGSEGV signal handler for shared memory misses
 * shared memory in this process.
 *
 * Overwritten SIGFAULT libc's handler. Tries to map the unmapped
 * shared memory, or will fail and call default libc's handler if
 * no new shared memory is mapped.
 * This is used only if the current system signal implementation
 * does not support the Realtime Signals Exception option or the
 * XSI Extension option.
 *
 * \param sigNo the signal id
 * \param sig_info the information associated with the signal
 * \param context the thread's context information where the signal was raised
 * \return nothing
 */
static void shm_sigsegv_sigaction(int sigNo, siginfo_t* sig_info, void* context)
{
	if (sigNo != SIGSEGV)
		return;

	if (!shm_update_mapped()) {
		/* no new shared mapping was set up
		 * we need to call libc's default handler
		 * WE ARE GOING TO CRASH!!! :)
		 * But first try to sync the shared memory of this process
		 */
		msync(shm, (size_t)shm->shm_end_slab - SLAB_SHARED_MAP_START, MS_SYNC | MS_INVALIDATE);
		if (libcSIGSEGV.sa_sigaction) {
			(libcSIGSEGV.sa_sigaction)(sigNo, sig_info, context);
		} else if (libcSIGSEGV.sa_handler) {
			libcSIGSEGV.sa_handler(sigNo);
		} else {
			/* we should never get here */
			abort();
		}
	}
}
/* ======================================================================================= */

#endif
