/*
 * 4PSA VoipNow - Implements a custom memory allocator based on the Memcached
 *                server implementation of a SLAB allocation.
 *                Our own implementation uses mmap and munmap for SLABs and
 *                internal dynamic data manipulation, and SLOT overhead
 *                logic for SLOT to SLAB class association (a SLOT overhead
 *                is a slot's additional data information added to each SLOT
 *                pointer address to be able to quickly identify it's parent
 *                SLAB class or SLAB address (only used for Valgrind profilling)
 *
 * Copyright (c) 2009 Rack-Soft, LLC. All rights reserved..
 *
 * $Header: //itsd/devel/platform/voipnow/251/low_level/kamailio_extra_sources/1.5.4/mem/slab_allocator.h#1 $
 *
 * Modified: $DateTime: 2010/09/28 18:16:02 $ by $Author: bogdanc $ with $Change: 71991 $
 */

#if !defined(SLAB_ALLOCATOR_H) && (defined(SLAB_MALLOC) || defined(SHM_SLAB_MALLOC))
#define SLAB_ALLOCATOR_H

#include <errno.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "meminfo.h"
#include "../config.h"
#include "../dprint.h"

/* ========================== Allocator's Implementation specifics ======================= */
#ifndef SLAB_MALLOC_F
#error "Must define an unique name for your SLAB malloc!\n"
#endif
#ifndef SLAB_FREE_F
#error "Must define an unique name for your SLAB free!\n"
#endif
#ifndef SLAB_REALLOC_F
#error "Must define an unique name for your SLAB realloc!\n"
#endif
#ifndef SLAB_INFO_F
#error "Must define an unique name for your SLAB info!\n"
#endif
#ifndef SLAB_STATUS_F
#error "Must define an unique name for your SLAB status!\n"
#endif

#ifndef SLAB_MAP_START
#error "Must define the address to start mapping memory */
#endif

#ifndef GET_SLAB_CLASS
#error "Must define a way to access each slab class for your SLAB allocator!\n"
#endif

#ifndef POWER_LAGEST
#error "Must define a way to access the largest slab class for your SLAB allocator!\n"
#endif

/*                          Defined by each SLAB allocator implementation                  */
static inline void*        internal_malloc (const size_t size);
static inline void         internal_free   (void *p, const size_t size);
static inline void*        internal_realloc(void *old_p, const size_t old_size, const size_t new_size);
/* ======================================================================================= */


/* ================================ SLAB sizing definitions  ============================= */
/*                                  USER defines - Kamailio                                */
#ifndef MAP_PREALLOC
#define MAP_PREALLOC 67108864 /* the size of the succesive mapped area to alloc SLABs from */
                              /* currently set to 64MB, meaning 64 SLABs                   */
#endif

#define POWER_SMALLEST 1    /* the minimum size of a SLOT's SLAB class (CHUNK_SIZE * 2^1)  */
#define POWER_LARGEST  200  /* the maximum size of a SLOT's SLAB class (CHUNK_SIZE * 2^199)*/
#ifndef POWER_BLOCK
#define POWER_BLOCK 1048576 /* the default size of a SLAB (1Mb) */
#endif
#define CHUNK_SIZE 128      /* first SLAB class SLOT size       */
/* ======================================================================================= */


/* ============================ SLAB ALLOCATOR Types and Data ============================ */
/* powers-of-N allocation structures to hold SLAB to SLOT information */
typedef struct _slabclass {
    unsigned int size;      /* sizes of items          */
    unsigned int perslab;   /* how many items per slab */

    void** slots;           /* list of item ptrs       */
    unsigned int sl_total;  /* size of previous array  */
    unsigned int sl_curr;   /* first free slot         */

    void* end_page_ptr;         /* pointer to next free item at end of page, or 0        */
    unsigned int end_page_free; /* number of items remaining at end of last alloced page */

    unsigned int slabs;     /* how many slabs were allocated for this class */

    void** slab_list;       /* array of SLAB pointers */
    unsigned int list_size; /* size of SLABs array    */
} slabclass_t;

#define MAX_NUMBER_OF_SLAB_CLASSES (POWER_LARGEST + 1) /* the array of SLAB classes size        */
#define SLAB_LIST_SIZE_PER_CLASS   511                 /* number of SLABs space in list         */
#define SLOTS_LIST_SIZE_PER_CLASS  511                 /* initial number of SLOTs space in list */
/* ======================================================================================= */


/* ================================= Forward Declarations ================================ */
/*                               Common Internal API for all SLABS                         */
static inline void         init_slabs      (const double factor);
static inline unsigned int slabs_clsid     (const size_t size);
static inline int          grow_slab_list  (const unsigned int id);
static        int          do_slabs_newslab(const unsigned int id);
/* ======================================================================================= */


/* ======================== SLAB OVERHEAD definitions - DO NOT TOUCH ===================== */
#define SLOT_OVERHEAD sizeof(void*) /* holds for each SLOT the SLAB class identifier       */

/* SLAB class(id) is defined as:
 *  [ MAGIC_COOKIE(31..16) | ID(15 .. 0) ] at (slot - SLOT_OVERHEAD) address */
#define SET_SLOT_SLABCLASS(slot, id) \
	*((size_t*) ((char*)slot - SLOT_OVERHEAD)) = \
		(size_t)( ((size_t)SLAB_MAGIC << 16) | ((size_t)id & 0xFFFF) );

#define GET_SLABCLASS_FROM_SLOT(slot) \
	*((size_t*) ((char*)slot - SLOT_OVERHEAD )) & 0xFFFF

#define IS_VALID_SLOT(slot) \
	(((*((size_t*) ((char*)slot - SLOT_OVERHEAD ))) >> 16) == (size_t)SLAB_MAGIC)


/* round a size to 4k page size */
#define POW_PAGE_SIZE 12  /* all mmap's size will be aligned at a page value (2^12 = 4k)   */
#define ROUNDTO_PAGE_SIZE(size) \
	((((size_t)(size) >> POW_PAGE_SIZE) + 1) << POW_PAGE_SIZE)

#define CHUNK_ALIGN_BYTES 8 /* SLOT allignment - 64 bits  */

#include "valgrind/valgrind_slab.h"   /* now we can include the Valgrind SLAB API          */
/* ======================================================================================= */


/* ============================== PRIVATE API implementation ============================= */
/**
 * \brief SLAB allocator system initialization
 *
 * Initializes the SLAB classes nedded by the SLAB allocator implementation
 * as well as internal private data. If prealloc is not used, each SLAB class
 * well have no SLAB allocated to it. It will be added as new objects of that
 * size are requested. One important parameter is the factor which indicated
 * the SLOT size in the SLABs classes. If a SLAB class has SLOT size x, then
 * the next SLAB class in the array will have a SLOT size = x * factor.
 * The factor is very important for memory usage. If your application has
 * multiple object of the same size, the factor can be big (2.0) as this will
 * lead to a greater amount of objects to be in the same SLAB class, which
 * will improve performance and memory usage(fragmantation).
 * This is the case for Kamailio.
 *
 * \param factor the factor used to grow the SLOT size in each new SLAB class
 * \return nothing
 */
static inline void init_slabs(const double factor)
{
	int i = POWER_SMALLEST - 1;
	unsigned int size = CHUNK_SIZE;

    /* Factor of 2.0 means use the default behavior */
	if (factor == 2.0 && size < 128)
		size = 128;


	slabclass_t* p = GET_SLAB_CLASS(0);

	while (++i < POWER_LARGEST && size <= POWER_BLOCK / 2) {
		/* Make sure items are always n-byte aligned */
		if (size % CHUNK_ALIGN_BYTES)
			size += CHUNK_ALIGN_BYTES - (size % CHUNK_ALIGN_BYTES);

		/* initialize current class */
		p->size    = size;
		p->perslab = POWER_BLOCK / size;
		p = GET_SLAB_CLASS(i);

		size *= factor;
#ifdef SLABS_MALLOC_DEBUG
        LM_DBG("slab class %3d: chunk size %6u perslab %5u\n",
                    i, slabclass[i].size, slabclass[i].perslab);
#endif
	}

	POWER_LAGEST = i;

	p->size    = POWER_BLOCK;
	p->perslab = 1;

}


/**
 * \brief Get's a SLAB class identifier given an object size
 *
 * A SLAB class is identified by the object's maximum size that can fill a free
 * SLOT in a SLAB type of the class. The functions returs the corresponding
 * SLAB class with the smallest SLOT size that can be used to hold the object.
 *
 * \param size the object's size
 * \return returns the class identifier (1 based), or 0 if no SLAB class is found
 */
static inline unsigned int slabs_clsid(const size_t size)
{
	int res        = 0;
	slabclass_t *p = GET_SLAB_CLASS(0);

	if (!p || !size)
		return 0;

    /* possibly an element will fit a smaller value */
	while (size > p->size) {
		if (++res >= POWER_LAGEST) {/* won't fit in the biggest SLAB */
			LM_ERR("Size %lu exceeds maximum shared memory SLAB\n",
					(unsigned long int)size);
			return 0;
		}
		p = GET_SLAB_CLASS(res); /* check next SLAB class */
	}
	return res + 1;
}


/**
 * \brief Increases the space for SLAB list in a SLAB class
 *
 * By default the initial size of the SLAB list is 0. When
 * the first SLAB will be allocated to a SLAB class,
 * the SLAB list will be increased to SLAB_LIST_SIZE_PER_CLASS
 * The internal memory to hold the list will be alloced using
 * a mmap'ed memory page. When the list will be filed
 * (not likely as 512 * 1Mb = 512Mb of heap allocated to the class)
 * the size of the list will double. We set 512 SLABS because
 * mmap's implementation will round the allocated memory to a page
 * in any case (512 * 8 = 4k of memory).
 *
 * \param id the SLAB class identifier
 * \return 1/-1 on SUCCESS/FAILUARE
 */
static inline int grow_slab_list(const unsigned int id)
{
    slabclass_t *p = GET_SLAB_CLASS(id - 1);
    if (p->slabs == p->list_size) {
    	void *new_list  = (void*)0;
    	size_t new_size = (p->list_size != 0) ? p->list_size * 2 : SLAB_LIST_SIZE_PER_CLASS;

    	new_list = internal_realloc(p->slab_list, p->list_size * sizeof(void*), new_size * sizeof(void*));

    	if (new_list == (void*)0)
    		return -1;

    	p->list_size = new_size;
    	p->slab_list = new_list;
    }
    return 1;
}


/**
 * \brief Allocates a new SLAB to a SLAB class to have additional SLOTS
 *
 * By default the initial size of new SLAB is either POWER_BLOCK
 * or the number of slots times slot size in the class + the number
 * and size of each slot's overhead information. The function will
 * grow the slab list of the SLAB class if needed and allocate the
 * heap memory to hold the SLAB(using mmap). It will also initialize
 * the overhead information for each SLOT in the newly allocated SLAB
 * and update the SLAB class's SLAB and SLOT infomation.
 *
 * \param id the SLAB class identifier
 * \return 1/-1 on SUCCESS/FAILUARE
 */
static int do_slabs_newslab(const unsigned int id)
{
	slabclass_t *p = GET_SLAB_CLASS(id - 1);
	/* alocate space for slots and slot slab information (unsigned long) */
#ifdef ALLOW_SLABS_REASSIGN
	int len = POWER_BLOCK + slabs[0].perslab * SLOT_OVERHEAD;
#else
	int len = p->perslab * (p->size + SLOT_OVERHEAD);
#endif
	char *ptr, *end;
	ptr = end = 0;

	if ( (grow_slab_list(id) <= 0) ||
		((ptr = internal_malloc((size_t)len)) == 0) ) {
		LM_ERR("Could not allocate any more slabs\n");
		return -1;
	}
	end = ptr + len;

	/* add the new slab memory pool to Valgrind's engine */
	PROCESS_VALGRIND_SLAB_OP(SLAB_NEW, ptr, ptr, len)

	/* add the new slab to the list */
	p->slab_list[p->slabs++] = ptr;

	/* intialize all slots */
	ptr += SLOT_OVERHEAD;
	p->end_page_ptr  = ptr;
	p->end_page_free = p->perslab;

	while (ptr < end) {
		/* initialize each slot with a abck reference to it's
		 * slab class and slab address(if __VALGRIND__ is defined) */
		SET_SLOT_SLABCLASS(ptr, id);
		SET_SLOT_SLABADDR(ptr, p->end_page_ptr - SLOT_OVERHEAD);
		ptr += p->size + SLOT_OVERHEAD;
	}

#ifdef SLABS_MALLOC_DEBUG
	LM_DBG("Alocated new slab for slab class %3d\n", id);
#endif

	return 1;
}
/* ======================================================================================= */


/* ============================== PUBLIC API implementation ============================== */
/**
 * \brief SLAB allocator implementation of malloc
 *
 * In order to identify a correct SLAB class for the desired size
 * we'll use internal defined slabs_clsid functionality.
 * Once a SLAB class is identified we'll try to allocate a new SLOT
 * from the free SLOTs list. If no SLOT is free we'll create a new SLAB
 * in this SLAB class, add it to the class's SLAB list, and use a SLOT from
 * the newly allocated SLAB. If a new SLAB is already allocated and have
 * free SLOTs allocate a SLOT from that new SLAB. The SLOT's overhead
 * information that we'll need later to free the object has been added
 * in do_slabs_newslab internal functionality.
 *
 * \param size the object's size to allocate
 * \return new address/0 on SUCCESS/FAILURE
 */
void* SLAB_MALLOC_F(const size_t size)
{
	slabclass_t *p  = 0;
	void* ret       = 0;
	unsigned int id = slabs_clsid(size);

	if (!id) {
		if (size > 0) {
			errno = ENOMEM;
			return ret;
		}
		/* if size == 0 allocate from the first slab list */
		id = 1;
	}

	p = GET_SLAB_CLASS(id - 1);

try_again:
	/* fail unless we have space at the end of a recently allocated page,
	 * we have something on our freelist, or we could allocate a new page */
	if ( (p->sl_curr == 0) && (p->end_page_ptr == (void*)0) &&
			(do_slabs_newslab(id) <= 0) ) {
		/* We don't have more memory available */
		errno = ENOMEM;
		return (void*)0;
    }

	if (p->sl_curr != 0) {
    	/* return off our freelist */
    	ret = p->slots[--p->sl_curr];
    } else {
    	/* if we recently allocated a whole page, return from that */
    	ret = p->end_page_ptr;
    	if (--p->end_page_free != 0) {
    		p->end_page_ptr = (void*)((char*)p->end_page_ptr + p->size + SLOT_OVERHEAD);
    	} else {
    		p->end_page_ptr = (void*)0;
    	}
   	}

	/* inform Valgrind's engine of the new alloc */
	PROCESS_VALGRIND_SLAB_OP(SLAB_ALLOC, ret, ret, size)

	if ( !IS_VALID_SLOT(ret) ) {
		LM_ERR("Invalid slot to allocate: %p, slab class: %3d\n", ret, id);
		goto try_again;
	}

#ifdef SLABS_MALLOC_DEBUG
	LM_DBG("Alocated at %p size %ld in slab %3d of size %ld\n", ret, size, id, p->size);
#endif

	errno = 0;
	return ret;
}


/**
 * \brief SLAB allocator implementation of free
 *
 * In order to free the object first we need to identify it's
 * parent SLAB class. This is done by using the SLOT overheads address.
 * The SLOT address is actually the address of the object, as this is
 * the address returned by slab_malloc. Once the corresponding SLAB class
 * has been identified, the current SLOT will be added to the SLAB class
 * free SLOTs list, which is grown if additional size is required to hold
 * the new SLOT address.
 * If the SLOT's overhead information has been corrupted (MAGIC_COOKIE invalid)
 * or the passed address is not one of a SLOT no action is done.
 *
 * \param ptr the object's address
 * \return nothing
 */
void SLAB_FREE_F(void *ptr)
{
	slabclass_t *p  = 0;
	unsigned int id = 0;

	if (!ptr)
		return;

	if ( !IS_VALID_SLOT(ptr) ||
			!(id = GET_SLABCLASS_FROM_SLOT(ptr)) || (id > POWER_LAGEST) ) {
		LM_ERR("Invalid slot to free: %p, slab class: %3d\n", ptr, id);
#ifdef SLAB_DEBUG
		abort();
#endif
		return;
	}

    p = GET_SLAB_CLASS(id - 1);

	if (p->sl_curr == p->sl_total) { /* need more space on the free list */
		void *new_slots = (void*)0;
		int new_size    = (p->sl_total != 0) ? p->sl_total * 2 : SLOTS_LIST_SIZE_PER_CLASS;

		new_slots = internal_realloc(p->slots, p->sl_total * sizeof(void*),	new_size * sizeof(void*));

		if (new_slots == (void*)0) {
			errno = ENOMEM;
			return;
		}

		p->slots    = new_slots;
		p->sl_total = new_size;
	}
	p->slots[p->sl_curr++] = ptr;

	/* inform Valgrind's engine of the free */
	PROCESS_VALGRIND_SLAB_OP(SLAB_DELETE, ptr, ptr, 0)

	return;
}


/**
 * \brief SLAB allocator implementation of realloc
 *
 * In order to realloc the object first we need to identify it's
 * parent SLAB class. This is done by using the SLOT overheads address.
 * The SLOT address is actually the address of the object, as this is
 * the address returned by slab_malloc. Once the corresponding SLAB class
 * has been identified, if the desired size is smaller that the SLOT size
 * the same SLOT address is used. If it is greater a new SLOT will be assigned
 * in another SLAB class, the current SLOT infomation will be copied to the new
 * SLOT and the current SLOT will be slab_free'd and the new SLOT returned.
 * If the SLOT's overhead information has been corrupted (MAGIC_COOKIE invalid)
 * or the passed address is not one of a SLOT no action is done.
 * FIXME: what if the desired size reduces the object to a SLOT size in a
 * smaller SLAB class... (not really used in Kamailio as realloc is used to grow)
 *
 * \param ptr the current object's address
 * \param size the new desired size of the object
 * \return new(old) address/0 on SUCCESS/FAILURE
 */
void* SLAB_REALLOC_F(void *ptr, const size_t size)
{
	unsigned int id        = 0;
	unsigned int slot_size = 0;

	if (!ptr) {
		return SLAB_MALLOC_F(size);
	}

	if ( !IS_VALID_SLOT(ptr) ||
			!(id = GET_SLABCLASS_FROM_SLOT(ptr)) || (id > POWER_LAGEST) ) {
		LM_ERR("Invalid slot to realloc: %p, slab class: %3d\n", ptr, id);
		errno = ENOMEM;
#ifdef SLAB_DEBUG
		abort();
#endif
		return (void*)0;
	}

	slot_size = GET_SLAB_CLASS(id - 1)->size;

	if (size <= slot_size) {
		if (size == 0) {
			/* free the pointer */
			SLAB_FREE_F(ptr);
			ptr = (void*)0;
		} else {
			/* inform Valgrind's engine of the realloc */
			PROCESS_VALGRIND_SLAB_OP(SLAB_REALLOC, ptr, ptr, size)
		}

		errno = 0;
		return ptr;
	}
	/* we need to allocate into a new slab */
	void* new_ptr = SLAB_MALLOC_F(size);
	if (new_ptr) {
		/* first we must make the slot's memory addressable and defined */
		PROCESS_VALGRIND_SLAB_OP(SLAB_REALLOC, ptr, ptr, slot_size)
		/* now we can safely copy the entire slot without any Valgrind's warnings */
		memcpy((char*)new_ptr, (char*)ptr, slot_size);
	}
	SLAB_FREE_F(ptr);

	return new_ptr;
}


/**
 * \brief Kamailio implemented functionality for dumping global SLAB allocator information
 *
 * Fills the information structure with data of the available total heap
 * memory to allocated, the used memory, the minimun size of a frag and
 * the number of allocated SLOTs in every class as well as the total OVERHEAD
 * size used for all the allocated SLOTs.
 *
 * \param info passed Kamailio defined information object
 * \return nothing
 */
void SLAB_INFO_F(struct mem_info* info)
{
	int it;
	unsigned long used_overhead = 0;
	unsigned long total_used    = 0;
	unsigned long total_free    = 0;
	unsigned long no_slots      = 0;

	if (!info)
		return;

	memset(info, 0, sizeof(*info));

	for(it = 0; it < MAX_NUMBER_OF_SLAB_CLASSES; it++){
		slabclass_t *p     = GET_SLAB_CLASS(it);
		info->total_frags += p->slabs;
		no_slots           = p->sl_curr + p->end_page_free;
		total_free        += no_slots * p->size;
		no_slots           = p->slabs * p->perslab - no_slots;
		used_overhead     += no_slots * SLOT_OVERHEAD;
		total_used        += no_slots * p->size;
	}

	info->min_frag    = POWER_BLOCK;
	info->total_size  = total_used;
	info->free        = total_free;
	info->max_used    = total_used + used_overhead;
	info->real_used   = (size_t)(prv_end_slab - SLAB_MAP_START);
}


/**
 * \brief Kamailio implemented functionality for dumping SLAB classes information
 *
 * Dumps the allocated and free SLOT information in each SLAB class that
 * have a SLAB allocated. Also dumps the number of SLOTs in each SLAB,
 * the allocated and free heap information for each SLAB class,
 * as well as relevant global SLAB allocator information.
 *
 * \return nothing
 */
void SLAB_STATUS_F(void)
{
	int i;
	struct mem_info info;
	unsigned long total = 0;

	SLAB_INFO_F(&info);

	LM_GEN1(memlog, "SLABS status\n");
	LM_GEN1(memlog, " heap size= %lu\n", info.real_used);
	LM_GEN1(memlog, " used= %lu, used+overhead=%lu, free=%lu\n",
			info.used, info.real_used, info.free);
	LM_GEN1(memlog, " max used (+overhead)= %lu\n", info.max_used);

	LM_GEN1(memlog, "dumping all slabclasses status:\n");
	for(i = 0; i < POWER_LARGEST; i++) {
		slabclass_t *p = GET_SLAB_CLASS(i);
		if (p->slabs == 0)
			continue;
		LM_GEN1(memlog, ">SLAB class id:%3d",    i);
		LM_GEN1(memlog, "  chunk_size: %u",      p->size);
		LM_GEN1(memlog, "  chunks_per_page: %u", p->perslab);
		LM_GEN1(memlog, "  total_pages: %u",     p->slabs);
		LM_GEN1(memlog, "  total_chunks: %u",    p->slabs * p->perslab);
		LM_GEN1(memlog, "  used_chunks: %u",     p->slabs * p->perslab - p->sl_curr - p->end_page_free);
		LM_GEN1(memlog, "  free_chunks: %u",     p->sl_curr + p->end_page_free);
		LM_GEN1(memlog, "  free_chunks_end: %u", p->end_page_free);

		total++;
	}

	/* add overall slab stats and append terminator */
	LM_GEN1(memlog, "\nactive_slab_class: %ld", total);
	LM_GEN1(memlog, "total_malloced: %llu", (unsigned long long)info.total_size);
	LM_GEN1(memlog, "-----------------------------\n");
}
/* ======================================================================================= */

#endif /* !SLAB_ALLOCATOR_H && (SLAB_MALLOC || SHM_SLAB_MALLOC) */
