/*
 * 4PSA VoipNow - Implements a custom memory allocator based on the Memcached
 *                server implementation of a SLAB allocation.
 *                Our own implementation uses mmap and munmap for SLABs and
 *                internal dynamic data manipulation, and SLOT overhead
 *                logic for SLOT to SLAB class association (a SLOT overhead
 *                is a slot's additional data information added to each SLOT
 *                pointer address to be able to quickly identify it's parent
 *                SLAB class or SLAB address (only used for Valgrind profilling)
 *
 * Copyright (c) 2009 Rack-Soft, LLC. All rights reserved..
 *
 * $Header: //itsd/devel/platform/voipnow/251/low_level/kamailio_extra_sources/1.5.4/mem/slab_malloc.c#1 $
 *
 * Modified: $DateTime: 2010/09/28 18:16:02 $ by $Author: bogdanc $ with $Change: 71991 $
 */

#ifdef SLAB_MALLOC

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "slab_malloc.h"
#include "../globals.h"
#include "../statistics.h"

/* =============================== Internal Types and Data =============================== */
struct _slabclass;
/* internal SLAB allocation data */
static struct _slabclass* slabclass;
static int power_largest   = 0;

/* internal variables for mapping private memory */
static void* prv_new_slab  = (void*)0; /* where mapped area starts from which to allocate a new SLAB  */
static void* prv_end_slab  = (void*)0; /* where mapped area ends and we need to map additional memory */
/* ======================================================================================= */


/* ====================== Private memory SLAB specifics definitions ====================== */
#define SLAB_MAP_START SLAB_PRIVATE_MAP_START

#define GET_SLAB_CLASS(class_id) (&slabclass[class_id])

#define POWER_LAGEST power_largest

#define SLAB_MALLOC_F  slabs_malloc
#define SLAB_FREE_F    slabs_free
#define SLAB_REALLOC_F slabs_realloc
#define SLAB_INFO_F    slabs_info
#define SLAB_STATUS_F  slabs_status

static size_t SLAB_MAGIC = 0xAAAA; /* A stands for privAte */

/* now we can safetly include the SLAB ALLOCATOR API */
#include "slab_allocator.h"

/* improve the locality of slab logic by statically allocating memory */
static slabclass_t mem_slabclass[MAX_NUMBER_OF_SLAB_CLASSES];
/* ======================================================================================= */


/* ================== Internal Macros for Private memory SLAB allocation ================= */
static inline void* internal_malloc(const size_t size)
{
	void *p = (void*)0;
	if ((size_t)(prv_end_slab - prv_new_slab) < size) {
		void *new_chunk = (void*)0;
		/* we need to map additional memory. We do it at a succesize address */
		if ( ( (new_chunk = mmap(prv_end_slab, MAP_PREALLOC, PROT_READ | PROT_WRITE,
					MAP_ANONYMOUS | MAP_PRIVATE | MAP_FIXED , -1, 0)) == (void*)-1 )
			|| (new_chunk != prv_end_slab) ) {
			LM_CRIT("Could not add additional private memory: mmap failed: %s\n", strerror(errno));
			if ( (new_chunk != (void*)-1) && (munmap(new_chunk, MAP_PREALLOC) == -1) ) {
				LM_ERR("Failed to unmap private memory at the wrong address: %s\n", strerror(errno));
			}
		} else {
			p            = prv_new_slab;
			prv_new_slab = (char*)prv_new_slab + size;
			prv_end_slab = (char*)prv_end_slab + MAP_PREALLOC;
			LM_NOTICE("Increased private memory with %lu MB from %p to %p",
					(unsigned long int)MAP_PREALLOC >> 20,
					(void*)(size_t)SLAB_PRIVATE_MAP_START,
					(void*)prv_end_slab);
		}
	} else {
		p            = prv_new_slab;
		prv_new_slab = (char*)prv_new_slab + size;
	}

	return p;
}


static inline void  internal_free(void *p, const size_t size)
{
	/* We currently do not support unmaping SLABs from memory */
}

static inline void* internal_realloc(void *old_p, const size_t old_size, const size_t new_size)
{
	void *new_p = internal_malloc(new_size);
	if (new_p && old_p) {
		memcpy(new_p, old_p, old_size);
		internal_free(old_p, old_size);
	}
	return new_p;
}
/* ======================================================================================= */


/* ============================== PUBLIC API implementation ============================== */
/**
 * \brief Private memory SLAB allocator system initialization
 *
 * Prior to calling the default initialization for the SLAB classes of a
 * SLAB allocator we initialize the variables for current range of the
 * available privately memory mapped. The range is currently 0 and starts
 * at SLAB_PRIVATE_MAP_START configuration option.
 *
 * \param factor the factor used to grow the SLOT size in each new SLAB class
 * \return nothing
 */
void slabs_init(const double factor)
{
	/* We are setting the private memory start and end address */
	prv_new_slab = prv_end_slab = (void*)(size_t)SLAB_PRIVATE_MAP_START;

	slabclass = mem_slabclass;

	memset(slabclass, 0, sizeof(slabclass_t) * MAX_NUMBER_OF_SLAB_CLASSES);

	/* we call the private slab allocation api to initialize the slabs */
	init_slabs(factor);
}
/* ======================================================================================= */

#endif
