/*
 * 4PSA VoipNow -  Valgrind API wrappers for profilling and debugging custom
 *                 SLAB based memory allocators. Each new SLAB in a SLAB class
 *                 is assigned a custom Valgrind memory pool, so Valgrind's
 *                 engine can follow the allocation and deallocation in this pool.
 *                 The pool identifier is the SLAB address and is held for each
 *                 SLOT at slot's address - sizeof(void*), so we can identify
 *                 from which SLAB the SLOT originated. In order to use the API
 *                 one must define a SLAB allocator implementation and __VALGRIND__
 *
 * Copyright (c) 2009 Rack-Soft, LLC. All rights reserved..
 *
 * $Header: //itsd/devel/platform/voipnow/251/low_level/kamailio_extra_sources/1.5.4/mem/valgrind/valgrind_slab.h#1 $
 *
 * Modified: $DateTime: 2010/09/28 18:16:02 $ by $Author: bogdanc $ with $Change: 71991 $
 */

#if !defined(VALGRIND_SLAB_H) && (defined(SLAB_MALLOC) || defined(SHM_SLAB_MALLOC))
#define VALGRIND_SLAB_H

#ifndef __VALGRIND__

/* empty definitions */
#define SET_SLOT_SLABADDR(slot, slab_addr)
#define GET_SLABADDR_FROM_SLOT(slot)
#define PROCESS_VALGRIND_SLAB_OP(type, old_addr, new_addr, size)
#define VALGRIND_SIGNAL_HANDLER

#else /* __VALGRIND__ */

#include "valgrind.h"
#include "memcheck.h"

#ifdef SLOT_OVERHEAD
	/* add Valgrind's SLAB address to a SLOT's overhead */
	static size_t original_overhead = (size_t)SLOT_OVERHEAD;
	static size_t valgrind_overhead = (size_t)SLOT_OVERHEAD + sizeof(void*);
	#undef  SLOT_OVERHEAD
	#define SLOT_OVERHEAD valgrind_overhead
#else
	#error "You must define SLOT_OVERHEAD before including this file!"
#endif

/* types of slab memory operation */
typedef enum {
	SLAB_NOP     = 0x0,
	SLAB_NEW     = 0x1,
	SLAB_ALLOC   = 0x2,
	SLAB_REALLOC = 0x3,
	SLAB_DELETE  = 0x4
}valgrind_slab_op_t;

/* SLAB address(slab_addr) is defined at
 * (slot - original_overhead) address
 * size_t type size for 64 bit systems */
#define SET_SLOT_SLABADDR(slot, slab_addr) \
	*((size_t*) ((char*)slot - original_overhead) ) = (size_t)slab_addr;
#define GET_SLABADDR_FROM_SLOT(slot) \
	*((size_t*) ((char*)slot - original_overhead) )

/* forward definitions in case we need to inform other valgrinds of our allocation */
#ifdef __VALGRIND_SHARED__
static inline void VALGRIND_INFORM(valgrind_slab_op_t type, void* old_addr, void* new_addr, int size);

#else
#define VALGRIND_INFORM(type,old_addr, new_addr, size)
#endif

/* inform Valgrind engine of an allocation or initialization of a memory pool
 * we take care of setting SLOT_OVERHEAD's to received addresses and sizes to
 * have a transparent API that the custom SLAB allocator can use. We add the
 * SLOT_OVERHEAD to the address and sizes in order not to be warned of accesing
 * those memory locations(they are used in the SLAB allocation logic) */
#define PROCESS_VALGRIND_SLAB_OP(type, old_addr, new_addr, size) \
	do { \
		switch (type) { \
		default: \
		case SLAB_NEW: \
			VALGRIND_CREATE_MEMPOOL(new_addr, 0, 0); \
			break; \
		case SLAB_ALLOC: { \
			/* first we must make this memory addressable and defined */ \
			VALGRIND_MAKE_MEM_DEFINED((char*)new_addr - SLOT_OVERHEAD, \
					SLOT_OVERHEAD); \
			size_t slab_addr = GET_SLABADDR_FROM_SLOT(new_addr); \
			/* now we can alloc in the correct valgrind mem pool */ \
			VALGRIND_MEMPOOL_ALLOC(slab_addr, \
				(char*)new_addr - SLOT_OVERHEAD, size + SLOT_OVERHEAD); \
			/* force reasigning slab class and address */ \
			VALGRIND_MAKE_MEM_DEFINED((char*)new_addr - SLOT_OVERHEAD, \
				SLOT_OVERHEAD); \
			break; \
		} \
		case SLAB_REALLOC: { \
			/* first we must make this memory addressable and defined */ \
			VALGRIND_MAKE_MEM_DEFINED((char*)old_addr - SLOT_OVERHEAD, \
					SLOT_OVERHEAD); \
			size_t slab_addr = GET_SLABADDR_FROM_SLOT(old_addr); \
			/* realloc using custom free and alloc */ \
			VALGRIND_MEMPOOL_FREE(slab_addr, \
				(char*)old_addr - SLOT_OVERHEAD); \
			VALGRIND_MEMPOOL_ALLOC(slab_addr, \
				(char*)new_addr - SLOT_OVERHEAD, size + SLOT_OVERHEAD); \
			/* force reasigning slab class and address */ \
			VALGRIND_MAKE_MEM_DEFINED((char*)new_addr - SLOT_OVERHEAD, \
				SLOT_OVERHEAD); \
			break; \
		} \
		case SLAB_DELETE: { \
			/* first we must make this memory addressable and defined */ \
			VALGRIND_MAKE_MEM_DEFINED((char*)new_addr - SLOT_OVERHEAD, \
				SLOT_OVERHEAD); \
			size_t slab_addr = GET_SLABADDR_FROM_SLOT(new_addr); \
			VALGRIND_MEMPOOL_FREE(slab_addr, \
					(char*)new_addr - SLOT_OVERHEAD); \
			break; \
		} \
		} \
		/* inform other valgrinds if needed */ \
		VALGRIND_INFORM(type, (void*)old_addr, (void*)new_addr, size); \
	} while(0);


#ifdef __VALGRIND_SHARED__
/* if a shared vlagrind is defined we must
 * inform all valgrinds of a process allocation
 */
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <signal.h>
#include <unistd.h>
#include "../../pt.h"

/* valgrind signal to use */
#define VALGRIND_SIG SIGUSR2
/* seconds to wait for all valgrinds to receive the signal */
#define VALGRIND_SEC_TIMEOUT 5

/* FIXME: forced system and Kamailio declarations */
#ifndef __USE_GNU
extern int semtimedop (int __semid, struct sembuf *__sops, size_t __nsops,
		       __const struct timespec *__timeout) __THROW;
#endif

/* structure to hold a shared memory operation */
typedef struct {
	volatile void*  old_addr;
	volatile void*  new_addr;
	volatile int    chunk_size;
	volatile valgrind_slab_op_t type;
}valgrind_op_t;
/* to store in shared memory the last made operation */
valgrind_op_t* last_shm_op;
/* semaphore used to process all the signals by child processes */
int* valgrind_sem;

/* wrapper to be used with malloc, free and realloc's custom allocators */
static inline void VALGRIND_INFORM(valgrind_slab_op_t type, void* old_addr, void* new_addr, int size)
{
	int it = 0;
	struct timespec ts = {0, 0};
	struct sembuf sop;
	sop.sem_num = 0;
	sop.sem_flg = 0;

	if (last_shm_op->type != SLAB_NOP) {
		/* try to avoid recursion */
		return;
	}

	if (!old_addr && (type == SLAB_REALLOC)) {
		type = SLAB_ALLOC; /* for first time allocation using realloc */
	}
	/* first set up the current shared memory operation */
	last_shm_op->old_addr   = old_addr;
	last_shm_op->new_addr   = new_addr;
	last_shm_op->chunk_size = size;
	last_shm_op->type       = type;

	/* set the semaphore value */
	if ( semctl(*valgrind_sem, 0, SETVAL, counted_processes) == -1) {
		LM_ERR("VALGRIND: Error on reseting semaphore errno:%d\n", (unsigned int)errno);
	}
	/* inform all valgrind process of our new change */
	sop.sem_op  = -1; /* decrement semaphore */
	pid_t current = getpid();
	for (; it < counted_processes; it++) {
		if (pt[it].pid && (pt[it].pid != current)) {
			kill(pt[it].pid, VALGRIND_SIG);
		} else if (semop(*valgrind_sem, &sop, 1) == -1) {
			LM_ERR("VALGRIND: Error on decrementing semaphore errno:%d\n", (unsigned int)errno);
		}
	}
	/* wait for all processes to receive the signal */
	sop.sem_op = 0; /* wait until it reaches 0 */
tryagain:
	ts.tv_sec = time(0) + VALGRIND_SEC_TIMEOUT;
	if (semtimedop(*valgrind_sem, &sop, 1, &ts) == -1) {
		if (errno==EINTR) {
			LM_DBG("VALGRIND: signal received while waiting for valgrinds to process the alloc\n");
			goto tryagain;
		}
		LM_ERR("VALGRIND: all other valgrinds may not have processed the alloc type:0x%x of addr:%p errno:%d\n",
				(unsigned int)type, new_addr, (unsigned int)errno);
	}
	/* now we can send additional signals */
	last_shm_op->type = SLAB_NOP;
}


static inline void* shared_valgrind_init(void* startAddr)
{
	last_shm_op   = (valgrind_op_t*)startAddr;
	startAddr     = (char*)startAddr + sizeof(valgrind_op_t);
	valgrind_sem  = (int*)startAddr;
	startAddr     = (char*)startAddr + sizeof(int);
	*valgrind_sem = semget(IPC_PRIVATE, 1, IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
	if (*valgrind_sem == -1) {
		LM_ERR("VALGRIND: Error on creating semaphore errno:%d\n", (unsigned int)errno);
	}

	last_shm_op->type = SLAB_NOP;
	return startAddr;
}


/* valgrind signal handler for processing another process allocation */
void valgrind_signal_handler()
{
	struct sembuf sop;
	sop.sem_num = 0;
	sop.sem_flg = 0;
	sop.sem_op  = -1; /* decrement semaphore */
	PROCESS_VALGRIND_SLAB_OP(last_shm_op->type,
		last_shm_op->old_addr, last_shm_op->new_addr, last_shm_op->chunk_size)
	if ( semop(*valgrind_sem, &sop, 1) == -1) {
		LM_ERR("VALGRIND: Error on decrementing semaphore errno:%d\n", (unsigned int)errno);
	}
}
#endif /* __VALGRIND_SHARED__ */

#endif /* __VALGRIND__ */

#endif /* !VALGRIND_SLAB_H && (SLAB_MALLOC || SHM_SLAB_MALLOC) */
