/*
 * $Id: //itsd/devel/platform/voipnow/251/low_level/kamailio_extra_sources/1.5.4/modules/load_balancer/load_balancer.c#1 $
 *
 * load balancer module - complex call load balancing
 *
 * Copyright (C) 2009 Voice Sistem SRL
 *
 * This file is part of opensips, a free SIP server.
 *
 * opensips is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * opensips is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * History:
 * --------
 *  2009-02-01 initial version (bogdan)
 */

#include "../../sr_module.h"
#include "../../db/db.h"
#include "../../dprint.h"
#include "../../error.h"
#include "../../ut.h"
#include "../../mod_fix.h"
#include "../../locking.h"
#include "../tm/tm_load.h"
//#include "../dialog/dlg_load.h"
#include "lb_parser.h"
#include "lb_db.h"
#include "lb_data.h"
#include "lb_prober.h"

MODULE_VERSION


/* db stuff */
static str db_url = str_init(DEFAULT_DB_URL);
static char *table_name = NULL;
gen_lock_t * g_sql_lock = NULL;

/* prober */
struct tm_binds tmb;
static int lb_ping_interval = 60;
/* number of consecutive outcomes after wich the state is changed */
int lb_probing_threshold = 10;
str lb_ping_method = str_init("OPTIONS");
str lb_ping_from   = str_init("sip:load_balancer@voipnow");

static int mod_init(void);
static int child_init(int rank);
static void mod_destroy(void);

static int fixup_resources(void** param, int param_no);

static int w_load_balance(struct sip_msg *req, char *rl);




static cmd_export_t cmds[]={
	{"load_balance", (cmd_function)w_load_balance,      1, fixup_resources,
			0, REQUEST_ROUTE|BRANCH_ROUTE|FAILURE_ROUTE},
	{0,0,0,0,0,0}
	};


static param_export_t mod_params[]={
	{ "db_url",                STR_PARAM, &db_url.s                 },
	{ "db_table",              STR_PARAM, &table_name               },
	{ "lb_ping_interval",		INT_PARAM, &lb_ping_interval},
	{ "lb_probing_threshold",		INT_PARAM, &lb_probing_threshold},
	{ 0,0,0 }
};


struct module_exports exports= {
	"load_balancer",  /* module's name */
	DEFAULT_DLFLAGS, /* dlopen flags */
	cmds,            /* exported functions */
	mod_params,      /* param exports */
	0,               /* exported statistics */
	0,         /* exported MI functions */
	0,               /* exported pseudo-variables */
	0,               /* extra processes */
	mod_init,        /* module initialization function */
	0,               /* reply processing function */
	mod_destroy,
	child_init       /* per-child init function */
};



static int fixup_resources(void** param, int param_no)
{
	struct lb_res_str_list *lb_rl;

	if (param_no==1) {
		/* parameter is string (semi-colon separated list) 
		 * of needed resources */
		lb_rl = parse_resorces_list( (char *)(*param), 0);
		if (lb_rl==NULL) {
			LM_ERR("invalid paramter %s\n",(char *)(*param));
			return E_CFG;
		}
		pkg_free(*param);
		*param = (void*)lb_rl;
	}

	return 0;
}


static int mod_init(void)
{
	int ret = -1;
	LM_INFO("Load-Balancer module - initializing\n");

	db_url.len = strlen(db_url.s);

	/* init and open DB connection */
	if (init_lb_db(&db_url, table_name)!=0) {
		LM_ERR("failed to initialize the DB support\n");
		goto _end;
	}

	g_sql_lock = lock_alloc();
	if (!g_sql_lock) {
		LM_CRIT("could not alloc lock\n");
		goto _end;
	}

	if (lock_init(g_sql_lock)==0){
		LM_CRIT("could not init lock\n");
		lock_dealloc((void*)g_sql_lock);
		g_sql_lock=0;
		goto _end;
	}

	g_node_list_lock = lock_alloc();
	if (!g_node_list_lock) {
		LM_CRIT("could not alloc lock\n");
		goto _end;
	}

	if (lock_init(g_node_list_lock)==0){
		LM_CRIT("could not init lock\n");
		lock_dealloc((void*)g_node_list_lock);
		g_node_list_lock=0;
		goto _end;
	}

	if (!(g_nodes = lb_node_list_init(0, 0, 0, 0, 0))) {
		LM_ERR("Not enough shared mem. init failed\n");
		goto _end;
	}

	/* Only, if the Probing-Timer is enabled the TM-API needs to be loaded: */
	if (lb_ping_interval > 0) {
		/*****************************************************
		 * TM-Bindings
	  	 *****************************************************/
		load_tm_f load_tm;
		load_tm=(load_tm_f)find_export("load_tm", 0, 0);
	
		/* import the TM auto-loading function */
		if (load_tm)
		{
			/* let the auto-loading function load all TM stuff */
			if (load_tm( &tmb ) == -1)
			{
				LM_ERR("could not load the TM-functions - disable DS ping\n");
				goto _end;
			}
			/*****************************************************
			 * Register the PING-Timer
	    	 *****************************************************/
			register_timer(lb_check_timer, NULL, lb_ping_interval);
		} else {
			LM_WARN("could not bind to the TM-Module, automatic"
					" re-activation disabled.\n");
		}
	}

	ret = 0;
_end:
	if (ret < 0) {
		if (g_node_list_lock) {
			lock_destroy(g_node_list_lock);
			lock_dealloc((void*)g_node_list_lock);
			g_node_list_lock=0;
		}
		if (g_sql_lock) {
			lock_destroy(g_sql_lock);
			lock_dealloc((void*)g_sql_lock);
			g_sql_lock=0;
		}
	}
	lb_close_db();

	return ret;
}


static int child_init(int rank)
{
	int ret = -1;

	/* init and open DB connection */
	if (init_lb_db(&db_url, table_name)!=0) {
		LM_ERR("failed to initialize the DB support\n");
		goto _end;
	}

	ret = 0;
_end:

	return ret;
}

static void mod_destroy(void)
{
	/* close DB connection */
	lb_close_db();

	//free node list
	
	if (g_node_list_lock) {
		lock_destroy(g_node_list_lock);
		lock_dealloc((void*)g_node_list_lock);
		g_node_list_lock=0;
	}
	if (g_sql_lock) {
		lock_destroy(g_sql_lock);
		lock_dealloc((void*)g_sql_lock);
		g_sql_lock=0;
	}
}


static int w_load_balance(struct sip_msg *req, char *rl)
{
	int ret;

	/* do lb */

	ret = do_load_balance(req, (struct lb_res_str_list*)rl);

	if (ret<0)
		return ret;
	return 1;
}
