/*
 * {app} - {purpose, description}
 *
 * Copyright (c) 2005-2006 Rack-Soft, LLC. All rights reserved.
 *
 * $Header: //itsd/devel/platform/voipnow/251/low_level/kamailio_extra_sources/1.5.4/modules/presence/presence_4psa.c#2 $
 *
 * Modified: $DateTime: 2010/10/27 19:46:48 $ by $Author: mihaim $ with $Change: 72773 $
 */



#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#include "../../pvar.h"
#include "../../mod_fix.h"
#include "../../db/db.h"
#include "../../sr_module.h"
#include "../../dprint.h"
#include "../../error.h"
#include "../../ut.h"
#include "../../parser/parse_to.h"
#include "../../parser/parse_uri.h" 
#include "../../parser/parse_content.h"
#include "../../parser/parse_from.h"
#include "../../mem/mem.h"
#include "../../mem/shm_mem.h"
#include "../../usr_avp.h"
#include "../tm/tm_load.h"
#include "../sl/sl_api.h"
#include "../../pt.h"
#include "../../mi/mi.h"
#include "../pua/hash.h"
#include "publish.h"
#include "subscribe.h"
#include "event_list.h"
#include "bind_presence.h"
#include "notify.h"
#include "pub_hash.h"
#include "../../db/db_utils_4psa.h"
#include "../domain/api.h"

static str ncomponent_table = str_init("node_component");
static str param_column     = str_init("param");
static str value_column     = str_init("value");
static str status_column    = str_init("status");

static str publisher_ip     = str_init("logical_publisher_server");
static str is_active        = str_init("1");

extern domain_api_t domainb;

const char *kPublishersEvent = "publish_servers";     /* mysql event in change_audit table for current publishers information version */
const char *kPublishersKey   = "sip.publish_servers"; /* redis key to check if the publishers information version has changed */

/**
 * \brief Loads publisher table to new hash table
 * This gets called either by mod_init or by a timer process only if the publish_servers event changed.
 * The new hash table gets loaded with information from node_component table with param=logical_publisher_server
 * and from ser_domain table with mwi_allow=1
 * In order to avoid any race conditions with the hash table switch an old copy of the information is also stored in memory
 * This means that any other process that started to iterate through the hash table will not crash because the old one
 * still exists in memory
 * 
 * \return 0/-1 on success/fail
 */
int load_publisher_table (void) {
	INIT_QUERY_WHERE(query, 2, 1);
	db_res_t* res = NULL;
	db_row_t* row;
	db_val_t* val;
	struct publisher_list **new_hash_table;
	int i, ret = -1;
	
	LM_INFO("Reloading accepted publisher servers from database\n");

	/* Choose new hash table and free its old contents */
	if (*hash_table == hash_table_1) {
		pub_hash_table_free(hash_table_2);
		new_hash_table = hash_table_2;
	} else {
		pub_hash_table_free(hash_table_1);
		new_hash_table = hash_table_1;
	}

	/* SELECT value FROM node_component */
	ADD_QUERY_RESULT_COLUMN(query, &value_column);
	/*    WHERE param=logical_publisher_server */
	ADD_QUERY_WHERE_COLUMN(query, OP_EQ, &param_column, STR, publisher_ip);
	/*       AND status=1 */
	ADD_QUERY_WHERE_COLUMN(query, OP_EQ, &status_column, STR, is_active);

	if (pa_dbf.use_table(pa_db, &ncomponent_table) < 0) {
		LM_ERR("Error while trying to use node_component table\n");
		goto try_ser_domain;
	}

	if (pa_dbf.query (pa_db, QUERY_COLS(query), QUERY_OPS(query), QUERY_VALS(query),
						QUERY_RESULTS(query), QUERY_COLS_CNT(query), QUERY_RESULTS_CNT(query), 0, &res) < 0) {
		LM_ERR("Error while querying database\n");
		goto try_ser_domain;
	}
	LM_DBG("Number of publisher-active nodeids found: %d\n", RES_ROW_N(res));

	row = RES_ROWS(res);
	for (i = 0; i < RES_ROW_N(res); i++) {
		val = ROW_VALUES(row + i);
		/* 4PSA: the value should be text (NOT nul term); which is of type DB_BLOB the same with blob */
		if ((ROW_N(row) == 1) && (VAL_TYPE(val) == DB_BLOB)) {
			char *big_data_copy = pkg_malloc(VAL_BLOB(val).len + 1);
			if (!big_data_copy) {
				LM_ERR("No more private memory\n");
				continue;
			}
			strncpy(big_data_copy, VAL_BLOB(val).s, VAL_BLOB(val).len);

			LM_DBG("Adding publisher server from node_component: %s \n", big_data_copy);

			if (pub_hash_table_install(new_hash_table, big_data_copy) == -1) {
				LM_ERR("Hash table error while creating publisher list\n");
			}
			pkg_free(big_data_copy);
		} else {
			LM_ERR("Database error while creating publisher list\n");
		}
	}
	ret = 0;

try_ser_domain:
	if (res) pa_dbf.free_result(pa_db, res);
	res = NULL;
	if (domainb.get_accepted_domains == 0)
		goto done;
	/* get the mwi publishers from ser_domain */
	if (!(res = domainb.get_accepted_domains(&pa_dbf, pa_db, 0/*mwi servers*/))) {
		LM_WARN("Error getting domain table. Could not reload accepted domains.");
		goto done;
	}

	row = RES_ROWS(res);
	for (i = 0; i < RES_ROW_N(res); i++) {
		val = ROW_VALUES(row + i);
		if ((ROW_N(row) == 1) && (VAL_TYPE(val) == DB_STRING)) {
			LM_DBG("Adding publisher server from ser_domain: %s \n", VAL_STRING(val));

			if (pub_hash_table_install(new_hash_table,(char*)VAL_STRING(val))==-1){
				LM_ERR("Hash table error while creating publisher list\n");
			}
		} else {
			LM_ERR("Database error while creating publisher list\n");
		}
	}

	ret = 0;
done:
	if (res) pa_dbf.free_result(pa_db, res);

	if (ret == 0) *hash_table = new_hash_table; /* if success switch tables */
	return ret;
}


/**
 * \brief Timer callback to be called on a timer process to update the publishers information
 * First checks if the publish_servers event changed an ONLY then updates the hash table information
 * by calling load_publisher_table.
 * publishers_id is a static variable to hold the current private value (table version) of the event
 *
 * \param ticks number of ticks passes - not used
 * \param param private data to be added to the call - not used
 */
void update_publishers(unsigned int ticks, void *param)
{
	static int publishers_id = 0; /* this timer routine will be run on a single process so we can have a static variable */

	if (check_set_event(g_pres_mem, &pa_dbf, pa_db, kPublishersEvent, kPublishersKey, &publishers_id)) {
		load_publisher_table();
	}
}


str * get_nat_local_contact(str local_ip, str local_port) {
	str * ret = NULL;
	str * con = NULL;

	con = pkg_malloc(sizeof(str));
	if (!con) {
		LM_ERR("failed to allocate memory for local_ip contact\n");
		goto _end;
	}
	con->s = NULL;
	con->len = 4 + local_ip.len + 1 + local_port.len;
	con->s = pkg_malloc(con->len + 1);
	if (!con->s) {
		LM_ERR("failed to allocate memory for local_ip contact\n");
		goto _end;
	}
	con->s[con->len] = '\0';

	snprintf(con->s, con->len+1, "sip:%.*s:%.*s", local_ip.len, local_ip.s, local_port.len, local_port.s);

	ret = con;
_end:
	if (!ret && con) {
		if (con->s) pkg_free(con->s);
		pkg_free(con);
	}

	return ret;
}


char * contact2dom(const char * str, size_t len, char * dest_ip, size_t max) {
	size_t pos = 0;

	if (!len && !(max-1)) return NULL;
	while (pos < len && str[pos++] != '@');
	if (pos == len) return NULL;

	str += pos;
	len -= pos;
	pos = 0;

	dest_ip[0] = '\0';
	while (pos < len && pos < max-1 && (str[pos] != ':' && str[pos] != ';' && str[pos] != '>')) {
		dest_ip[pos] = str[pos];
		++pos;
	}
	dest_ip[pos] = '\0';

	return dest_ip;
}


