/*
 * Monitoring of added/removed interfaces using libnl
 *
 * Copyright (c) 2005-2009 Rack-Soft, LLC. All rights reserved.
 *
 * $Header: //itsd/devel/platform/voipnow/251/low_level/kamailio_extra_sources/1.5.4/modules/usrloc/netlink_monitor_4psa.c#1 $
 *
 * Modified: $DateTime: 2010/09/28 18:16:02 $ by $Author: bogdanc $ with $Change: 71991 $
 */

#include <stdio.h>
#include <inttypes.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <poll.h>
#include <sys/socket.h>
#include <asm/types.h>

#include "netlink_monitor_4psa.h"


#include <netlink/netlink.h>
#include <netlink/addr.h>
#include <netlink/route/link.h>
#include <netlink/route/addr.h>
#include <netlink/cache.h>

static struct nl_handle *handle = NULL;
static struct nl_cache  *link_cache = NULL;

static void obj_input(struct nl_object *obj, void *arg);
static int event_input(struct nl_msg *msg, void *arg);
static int init_netlink_monitoring_nl(void);
static void stop_netlink_monitoring_nl(void);

static const enum rtnetlink_groups netlink_event_groups[] = {
		RTNLGRP_IPV4_IFADDR,
		RTNLGRP_IPV6_IFADDR,
		RTNLGRP_NONE
	};

int nl_fd = -1;

static void obj_input(struct nl_object *obj, void *arg)
{
}

static int event_input(struct nl_msg *msg, void *arg)
{
	if (nl_msg_parse(msg, &obj_input, NULL) < 0) {
		printf("NETLINK Unknown message type\n");
	}

	return NL_EXIT;
}

static int init_netlink_monitoring_nl(void)
{
	int i;
	int err;
	
	if (handle) {
		return nl_fd;
	}
	
	/* Allocates and initializes a new netlink handle, the netlink process id
	 * is set to the local process id which may conflict if multiple handles
	 * are created, therefore you may have to overwrite it using
	 * nl_handle_set_pid(). The initial sequence number is initialized to the
	 * current UNIX time. The custom callback (NL_CB_VERBOSE) handlers are
	 * being used.
	 */
	if ((handle = nl_handle_alloc()) == NULL)
		goto error;
	
	/* As this socket used for event processing we need to disable sequence
	 * number checking in order to be able to receive messages not explicitely
	 * requested.
	 */
	nl_disable_sequence_check(handle);
	
	/* set the callback for receiving and processing and NETLINK message */
	nl_cb_set(nl_handle_get_cb(handle), NL_CB_VALID, NL_CB_CUSTOM, event_input, NULL);
	
	/* We we'll join the fallowing NETLINK groups to catch events sent by the kernel
	 * - RTMGRP_LINK (network interface create/delete/up/down events) multicast group
	 */
	nl_join_groups(handle, RTMGRP_LINK);
	
	/* Bind and connect the socket to a protocol, NETLINK_ROUTE */
	if ((err = nl_connect(handle, NETLINK_ROUTE)) < 0) {
		printf("NETLINK connect error: %s\n", nl_geterror());
		goto error;
	}
	
	for (i = 0; netlink_event_groups[i] != RTNLGRP_NONE; i++) {
		if ((err = nl_join_group(handle, netlink_event_groups[i])) < 0) {
			printf("NETLINK group join error: %s\n", nl_geterror());
			goto error;
		}
	}
	
	if ((link_cache = rtnl_link_alloc_cache(handle)) == NULL) {
		printf("NETLINK link cache error: %s\n", nl_geterror());
		goto error;
	}
	nl_cache_mngt_provide(link_cache);
	
	/* Returns the file handle we can use to poll */
	nl_fd = nl_handle_get_fd(handle);
	return nl_fd;

error:
	stop_netlink_monitoring_nl();
	return -1;
}

static void stop_netlink_monitoring_nl(void)
{
	if (link_cache) {
		nl_cache_clear(link_cache);
		nl_cache_free(link_cache);
		link_cache = NULL;
	}
	if (handle) {
		nl_close(handle);
		nl_handle_destroy(handle);
		handle = NULL;
	}
}

int test_netlink_settup()
{
	int isOk = (init_netlink_monitoring_nl() == -1)? 0 : 1;
	stop_netlink_monitoring_nl();
	
	return isOk;
}

int wait_for_change_nl()
{
	int rc;
	struct pollfd fds[1];
	int nfds = 1;
	
	if (init_netlink_monitoring_nl() == -1) {
		return -1;
	}
	
	fds[0].fd = nl_fd;
	fds[0].events = POLLIN;

_again:
	/* wait for an incoming message on the netlink socket */
	rc = poll(fds, nfds, -1);	
	
	if (rc < 0) {
		if (errno == EINTR) {
			goto _again;
		}
		printf("Error in poll on netlink socket: %s.\n", strerror(errno));
		return -1;	
	}
	
	/* Calls the NETLINK library receive message handler */
	nl_recvmsgs_def(handle);
	
	nl_cache_clear(link_cache);
	
	return 0;
}

