/*
 * $Id: script_cb.c 5916 2009-08-31 10:31:40Z henningw $
 *
 * Copyright (C) 2001-2003 FhG Fokus
 *
 * This file is part of Kamailio, a free SIP server.
 *
 * Kamailio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * Kamailio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * History:
 * --------
 *  2003-03-29  cleaning pkg allocation introduced (jiri)
 *  2003-03-19  replaced all mallocs/frees w/ pkg_malloc/pkg_free (andrei)
 *  2005-02-13  script callbacks devided into request and reply types (bogdan)
 */

/*!
 * \file
 * \brief Script callbacks functions
 */


#include <stdlib.h>
#include "script_cb.h"
#include "dprint.h"
#include "error.h"
#include "mem/mem.h"

/*! list of pre-request callbacks */
static struct script_cb *pre_req_cb=0;
/*! list of post-request callbacks */
static struct script_cb *post_req_cb=0;

/*! list of pre-reply callbacks */
static struct script_cb *pre_rpl_cb=0;
/*! list of post-reply callbacks */
static struct script_cb *post_rpl_cb=0;

/*! global callback id counter */
static unsigned int cb_id=0;


/*!
 * \brief Add a callback to the list
 * \param list list of callbacks
 * \param f 
 * \param param
 * \return 0 on success, -1 on error
 */
static inline int add_callback( struct script_cb **list,
	cb_function f, void *param)
{
	struct script_cb *new_cb;

	new_cb=pkg_malloc(sizeof(struct script_cb));
	if (new_cb==0) {
		LM_ERR("out of pkg memory\n");
		return -1;
	}
	new_cb->cbf = f;
	new_cb->id = cb_id++;
	new_cb->param = param;
	/* link at the beginning of the list */
	new_cb->next = *list;
	*list = new_cb;
	return 0;
}


/*!
 * \brief Register a script callback function
 * \param f callback function
 * \param type callback type
 * \param param callback parameter
 * \return 0 on success, -1 on error
 */
int register_script_cb( cb_function f, int type, void *param )
{
	/* type checkings */
	if ( (type&(REQ_TYPE_CB|RPL_TYPE_CB))==0 ) {
		LM_CRIT("invalid callback, no request or reply type specified\n");
		return -1;
	}
	if ( (type&(PRE_SCRIPT_CB|POST_SCRIPT_CB))==0 ||
	(type&PRE_SCRIPT_CB && type&POST_SCRIPT_CB) ) {
		LM_CRIT("invalid callback, please set either the POST or PRE type\n");
		return -1;
	}

	if (type&REQ_TYPE_CB) {
		/* callback for request script */
		if (type&PRE_SCRIPT_CB) {
			if (add_callback( &pre_req_cb, f, param)<0)
				goto error;
		} else if (type&POST_SCRIPT_CB) {
			if (add_callback( &post_req_cb, f, param)<0)
				goto error;
		}
	}
	if (type&RPL_TYPE_CB) {
		/* callback (also) for reply script */
		if (type&PRE_SCRIPT_CB) {
			if (add_callback( &pre_rpl_cb, f, param)<0)
				goto error;
		} else if (type&POST_SCRIPT_CB) {
			if (add_callback( &post_rpl_cb, f, param)<0)
				goto error;
		}
	}

	return 0;
error:
	LM_ERR("failed to add callback\n");
	return -1;
}


/*!
 * \brief Destroy callback list, free memory
 * \param list callback list
 */
static inline void destroy_cb_list(struct script_cb **list)
{
	struct script_cb *foo;

	while( *list ) {
		foo = *list;
		*list = (*list)->next;
		pkg_free( foo );
	}
}


/*!
 * \brief Destroy all callbacks
 */
void destroy_script_cb(void)
{
	destroy_cb_list( &pre_req_cb  );
	destroy_cb_list( &post_req_cb );
	destroy_cb_list( &pre_rpl_cb  );
	destroy_cb_list( &post_rpl_cb );
}


/*!
 * \brief Execute pre-script callbacks, stop on error
 * \param msg SIP message
 * \param cb callback list
 * \return 1 on success, 0 on error
 */
static inline int exec_pre_cb( struct sip_msg *msg, struct script_cb *cb)
{
	for ( ; cb ; cb=cb->next ) {
		/* stop on error */
		if (cb->cbf(msg, cb->param)==0)
			return 0;
	}
	return 1;
}


/*!
 * \brief Execute post-script callbacks, don't stop on error
 * \param msg SIP message
 * \param cb callback list
 * \return 1
 */
static inline int exec_post_cb( struct sip_msg *msg, struct script_cb *cb)
{
	for ( ; cb ; cb=cb->next){
		cb->cbf( msg, cb->param);
	}
	return 1;
}


/*!
 * \brief Execute pre-request callbacks
 * \param msg SIP message
 * \return 1 on success, 0 on error
 */
int exec_pre_req_cb( struct sip_msg *msg)
{
	return exec_pre_cb( msg, pre_req_cb);
}


/*!
 * \brief Execute pre-reply callbacks
 * \param msg SIP message
 * \return 1 on success, 0 on error
 */
int exec_pre_rpl_cb( struct sip_msg *msg)
{
	return exec_pre_cb( msg, pre_rpl_cb);
}


/*!
 * \brief Execute post-request callbacks
 * \param msg SIP message
 * \return 1
 */
int exec_post_req_cb( struct sip_msg *msg)
{
	return exec_post_cb( msg, post_req_cb);
}


/*!
 * \brief Execute post-reply callbacks
 * \param msg SIP message
 * \return 1
 */
int exec_post_rpl_cb( struct sip_msg *msg)
{
	return exec_post_cb( msg, post_rpl_cb);
}
