/*
 * $Id: usr_avp.h 5299 2008-12-04 18:12:33Z henningw $
 *
 * Copyright (C) 2001-2003 FhG Fokus
 *
 * This file is part of Kamailio, a free SIP server.
 *
 * Kamailio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * Kamailio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * History:
 * ---------
 *  2004-07-21  created (bogdan)
 *  2004-11-14  global aliases support added (bogdan)
 *  2005-02-14  list with FLAGS USAGE added (bogdan)
 */


/*!
 * \file
 * \brief AVP handling functions
 * - \ref AVPFlagsText
 * \todo 1) int_str -> (int,str), 2) avp is double linked list (faster at delete and insert)
 */

/*!
 * \page AVPFlagsText AVP flags and their meaning and owner
 *
 * List with the allocated flags, their meaning and owner.
 * [0-7] - internal flags; [8-15] - to be used by script
 *  flag no.    owner            description
 * -------------------------------------------------------
 *   0        avp_core          avp has a string name
 *   1        avp_core          avp has a string value
 *   2        core              contact avp qvalue change
 *   7        avpops module     avp was loaded from DB
 */

#ifndef _SER_URS_AVP_H_
#define _SER_URS_AVP_H_


#include "str.h"


/*! basic data type for AVPs */
typedef union {
	int  n;
	str s;
} int_str;


/*! AVP values and list */
struct usr_avp {
	unsigned short id;
	unsigned short flags;
	struct usr_avp *next;
	void *data;
};


/*! delimiter between data type and name */
#define AVP_NAME_DELIM	':'

#define AVP_NAME_VALUE_MASK	0x0003
#define AVP_CORE_MASK		0x00ff
#define AVP_SCRIPT_MASK		0xff00
#define avp_core_flags(f)	((f)&0x00ff)
#define avp_script_flags(f)	(((f)<<8)&0xff00)
#define avp_get_script_flags(f)	(((f)&0xff00)>>8)

#define AVP_NAME_STR     (1<<0)
#define AVP_VAL_STR      (1<<1)

#define is_avp_str_name(a)	(a->flags&AVP_NAME_STR)
#define is_avp_str_val(a)	(a->flags&AVP_VAL_STR)


/*!
 * \brief Add an AVP to the global list
 * \param flags AVP flags
 * \param name AVP name
 * \param val AVP value
 * \return 0 on success, -1 on failure
 */
int add_avp( unsigned short flags, int_str name, int_str val);


/*!
 * \brief Search the first matching AVP in the global list
 *
 * Search the first AVP in the global list that matches to the given parameter.
 * \param flags AVP flags
 * \param name AVP name
 * \param val AVP value
 * \param start start the search at start->next, if NULL start from the head of the global list
 * \return found AVP, or NULL on failure
 */
struct usr_avp *search_first_avp( unsigned short flags, int_str name,
									int_str *val,  struct usr_avp *start);


/*!
 * \brief Search the next AVP in the given AVP list
 *
 * Search the next AVP in the given AVP list, that is returned from search_first_avp
 * \param avp AVP list
 * \param val AVP value
 * \return found AVP, or NULL on failure
 */
struct usr_avp *search_next_avp( struct usr_avp *avp, int_str *val  );


/*!
 * \brief Destroy global AVP list
 */
void reset_avps(void);


/*!
 * \brief Search AVP in list and delete it
 * \param avp searched AVP
 */
void destroy_avp( struct usr_avp *avp);


/*!
 * \brief Search AVP(s) in list and destroy them
 * \param flags AVP flags
 * \param name AVP name
 * \param all set to 1 to delete all matching AVPs, 0 to delete only the first
 * \return the number of AVPs that were deleted
 */
int destroy_avps( unsigned short flags, int_str name, int all);


/*!
 * \brief Free an AVP list in shared memory with memory locking
 * \param list AVP list
 */
void destroy_avp_list( struct usr_avp **list );


/*!
 * \brief Free an AVP list in shared memory without memory locking
 * \param list AVP list
 */
void destroy_avp_list_unsafe( struct usr_avp **list );


/*!
 * \brief Get AVP value
 * \param avp inspected AVP
 * \param val returned value
 */
void get_avp_val(struct usr_avp *avp, int_str *val );


/*!
 * \brief Get AVP name
 * \param avp inspected AVP
 * \return pointer to the name, 0 on error
 */
str* get_avp_name(struct usr_avp *avp);


/*!
 * \brief Set global AVP list, assert that old list is not null
 * \param list new AVP list
 * \return pointer to old list
 */
struct usr_avp** set_avp_list( struct usr_avp **list );


/*!
 * \brief Return global AVP list, assert that is not null
 * \return global AVP list
 */
struct usr_avp** get_avp_list(void);


/*!
 * \brief Add global alias 
 * \param alias_definition alias definition, including the AVP
 * \return 0 on success, -1 on error
 */
int add_avp_galias_str(char *alias_definition);


/*!
 * \brief Lookup global AVP alias
 * \param alias alias string
 * \param type AVP type
 * \param avp_name AVP name
 * \return 0 if alias to specified AVP could be found, -1 if not found
 */
int lookup_avp_galias(str *alias, int *type, int_str *avp_name);


/*!
 * \brief Add global AVP alias to the list, allocate private memory
 * \param alias alias string
 * \param type AVP type
 * \param avp_name AVP name
 * \return 0 on success, -1 on failure
 */
int add_avp_galias(str *alias, int type, int_str avp_name);


/*!
 * \brief Parse an AVP name (as string) and set type and name
 * \param name AVP name
 * \param type AVP type
 * \param avp_name AVP name
 * \return 0 on success, -1 on failure
 */
int parse_avp_name( str *name, int *type, int_str *avp_name);


/*!
 * \brief Parse AVP specification (alias or AVP) and set type and name
 * \param name AVP name
 * \param type AVP type
 * \param avp_name 
 * \return 0 on success, -1 on failure
 */
int parse_avp_spec( str *name, int *type, int_str *avp_name);

#endif
